from ..utils.events import EmitterGroup


class Camera:
    """Camera object modeling position and view of the camera.

    Parameters
    ----------
    dims : napari.components.Dims
        Dims model of the viewer.
    angles : 3-tuple
        Euler angles of camera in 3D viewing (rx, ry, rz), in degrees.
        Only used during 3D viewing.
    center : 2-tuple or 3-tuple
        Center of the camera for either 2D or 3D viewing.
    zoom : float
        Scale from canvas pixels to world pixels.

    Attributes
    ----------
    angles : 3-tuple
        Euler angles of camera in 3D viewing (rx, ry, rz), in degrees.
        Only used during 3D viewing.
    center : 2-tuple or 3-tuple
        Center of the camera for either 2D or 3D viewing.
    ndisplay : int
        Number of displayed dimensions, must be either 2 or 3.
    zoom : float
        Scale from canvas pixels to world pixels.
    """

    def __init__(self, dims, *, center=None, zoom=1, angles=(0, 0, 90)):
        self._dims = dims

        if center is None:
            center = (0,) * self._dims.ndisplay
        center = center[-self._dims.ndisplay :]

        self._zoom = zoom
        self._angles = angles
        self._center = center

        self.events = EmitterGroup(
            source=self,
            auto_connect=True,
            angles=None,
            zoom=None,
            center=None,
        )

    @property
    def center(self):
        """tuple: Center point of camera view for 2D or 3D viewing."""
        return self._center

    @center.setter
    def center(self, center):
        if self.center == tuple(center):
            return
        if self.ndisplay != len(center):
            raise ValueError(
                f'Center must be same length as currently displayed'
                f' dimensions, got {len(center)} need {self.ndisplay}.'
            )
        self._center = tuple(center)
        self.events.center()

    @property
    def ndisplay(self):
        """int: Dimensionality of the camera rendering."""
        return self._dims.ndisplay

    @ndisplay.setter
    def ndisplay(self, ndisplay):
        self._dims.ndisplay = ndisplay

    @property
    def zoom(self):
        """float: Scale from canvas pixels to world pixels."""
        return self._zoom

    @zoom.setter
    def zoom(self, zoom):
        if self._zoom == zoom:
            return
        self._zoom = zoom
        self.events.zoom()

    @property
    def angles(self):
        """3-tuple: Euler angles of camera in 3D viewing, in degrees."""
        if self.ndisplay == 3:
            return self._angles
        else:
            return (0, 0, 90)

    @angles.setter
    def angles(self, angles):
        if self._angles == tuple(angles):
            return
        self._angles = tuple(angles)
        self.events.angles()



"""QtTestImage class.

Creating test images is meant as an internal developer feature for now.
"""
from collections import namedtuple
from typing import Callable, Tuple

from qtpy.QtWidgets import (
    QComboBox,
    QFrame,
    QGroupBox,
    QLabel,
    QPushButton,
    QVBoxLayout,
)

from ....components.experimental.chunk import async_config
from ....layers.image.experimental import TestImageSettings
from ....utils import config
from .image_creator import create_test_image_multi
from .qt_render_widgets import QtLabeledComboBox, QtLabeledSpinBox

Callback = Callable[[], None]
IntCallback = Callable[[int], None]

TILE_SIZE_RANGE = range(1, 4096, 100)

IMAGE_SHAPE_DEFAULT = (1024, 1024)  # (height, width)
IMAGE_SHAPE_RANGE = range(1, 65536, 100)


# We can create 3 types of images layers.
IMAGE_TYPES = {
    "Normal": config.CREATE_IMAGE_NORMAL,
    "Compound": config.CREATE_IMAGE_COMPOUND,
    "Tiled": config.CREATE_IMAGE_TILED,
}
IMAGE_TYPE_DEFAULT = "Tiled"

# Allow the user to pick test images by name. If the "shape" is none then
# the user can chose the shape with two spin controls.
TEST_IMAGES = {
    "Digits": {
        "shape": None,
        "factory": lambda image_settings: create_test_image_multi(
            "0", image_settings
        ),
    },
}
TEST_IMAGE_DEFAULT = "Digits"

# Add skimage.data images if installed. Napari does not depend on
# skimage but many developers will have it.
try:
    import skimage.data

    TEST_IMAGES.update(
        {
            "Astronaut": {
                "shape": (512, 512),
                "factory": lambda: skimage.data.astronaut(),
            },
            "Chelsea": {
                "shape": (300, 451),
                "factory": lambda: skimage.data.chelsea(),
            },
            "Coffee": {
                "shape": (400, 600),
                "factory": lambda: skimage.data.coffee(),
            },
        }
    )
except ImportError:
    pass  # The skimage.data images won't be available.


class QtVariableShape(QGroupBox):
    """Two spin boxes used to set the shape of an image."""

    def __init__(self):
        super().__init__("Dimensions")

        layout = QVBoxLayout()
        self.height = QtLabeledSpinBox(
            "Height", IMAGE_SHAPE_DEFAULT[0], IMAGE_SHAPE_RANGE
        )
        self.width = QtLabeledSpinBox(
            "Width", IMAGE_SHAPE_DEFAULT[1], IMAGE_SHAPE_RANGE
        )
        layout.addWidget(self.height)
        layout.addWidget(self.width)
        self.setLayout(layout)

    def get_shape(self) -> Tuple[int, int]:
        """Return the currently configured shape.

        Return
        ------
        Tuple[int, int]
            The requested [height, width] shape.
        """
        return self.height.spin.value(), self.width.spin.value()


class QtFixedShape(QGroupBox):
    """A label to display the fixed shape of an image."""

    def __init__(self):
        super().__init__("Details")

        layout = QVBoxLayout()
        self.shape = QLabel("Shape: ???")
        layout.addWidget(self.shape)
        self.setLayout(layout)

    def set_shape(self, shape: Tuple[int, int]) -> None:
        """Set the shape to show in the labels.

        shape : Tuple[int, int]
            The shape to show in the labels.
        """
        self.shape.setText(f"Shape: ({shape[0]}, {shape[1]})")


class QtTestImageLayout(QVBoxLayout):
    """Controls to a create a new test image layer.

    Parameters
    ----------
    on_create : Callable[[], None]
        Called when the create test image button is pressed.
    """

    def __init__(self, on_create: Callback):
        super().__init__()
        self.addStretch(1)

        # Shows the test images we can create.
        self.name = QComboBox()
        self.name.addItems(TEST_IMAGES.keys())
        self.name.activated[str].connect(self._on_name)
        self.addWidget(self.name)

        # Create shape controls for "variable sized" or "fixed size" images.
        # Add both sets of controls, but only one set is visible at a time.
        ShapeControls = namedtuple('ShapeControls', "variable fixed")
        self.shape_controls = ShapeControls(QtVariableShape(), QtFixedShape())
        self.addWidget(self.shape_controls.variable)
        self.addWidget(self.shape_controls.fixed)

        # The tile size is available with either type of octree layer.
        self.tile_size = QtLabeledSpinBox(
            "Tile Size", async_config.octree.tile_size, TILE_SIZE_RANGE
        )
        self.addWidget(self.tile_size)

        # Which type of image layer/visual to create.
        self.image_type = QtLabeledComboBox("Type", IMAGE_TYPES)
        self.image_type.set_value(config.create_image_type)
        self.image_type.combo.activated[str].connect(self._on_type)
        self.addWidget(self.image_type)

        # The create button.
        button = QPushButton("Create Test Image")
        button.setToolTip("Create a new test image")
        button.clicked.connect(on_create)
        self.addWidget(button)

        # Set the initially selected image.
        self._on_name(TEST_IMAGE_DEFAULT)

    def _on_name(self, value: str) -> None:
        """Called when a new image name is selected.

        Set which image controls are visible based on the spec of
        the newly selected image.

        Parameters
        ----------
        value : str
            The new image name.
        """
        spec = TEST_IMAGES[value]

        if spec['shape'] is None:
            # Image has a settable shape.
            self.shape_controls.variable.show()
            self.shape_controls.fixed.hide()
        else:
            # Image has a fixed shape.
            self.shape_controls.variable.hide()
            self.shape_controls.fixed.show()
            self.shape_controls.fixed.set_shape(spec['shape'])

    def _on_type(self, _value: str) -> None:
        """User changed which type image they want to create.

        Parameters
        ----------
        value : str
            The new image type.
        """
        # Only show tile_size for octree-based images.
        self.tile_size.setVisible(
            self.image_type.get_value() != config.CREATE_IMAGE_NORMAL
        )

    @property
    def settings(self) -> TestImageSettings:
        """The desired image settings.

        Return
        ------
        ImageSetttings
            The desired image settings.
        """
        return TestImageSettings(
            self.shape_controls.variable.get_shape(),
            self.tile_size.spin.value(),
        )


class QtTestImage(QFrame):
    """Frame with controls to create a new test image.

    Parameters
    ----------
    viewer : Viewer
        The napari viewer.
    """

    # Class attribute so system-wide we create unique names, even if
    # created from different QtRender widgets for different layers.
    image_index = 0

    def __init__(self, viewer):
        super().__init__()
        self.viewer = viewer
        self.layout = QtTestImageLayout(self._create_test_image)
        self.setLayout(self.layout)

    def _create_test_image(self) -> None:
        """Create a new test image layer."""

        # Get the spec for the current selected type of image.
        image_name = self.layout.name.currentText()
        spec = TEST_IMAGES[image_name]
        factory = spec['factory']
        image_settings = self.layout.settings

        if spec['shape'] is None:
            # This image has a settable shape provided by the UI, so we
            # pass the settings into the factory.
            data = factory(image_settings)
        else:
            # This image comes in just one specific shape, so the factory
            # takes no arguments.
            data = factory()

        # Give each new layer a unique name.
        unique_name = f"test-image-{QtTestImage.image_index:003}"
        QtTestImage.image_index += 1

        # Set config for which type of image to create.
        image_type = self.layout.image_type.get_value()
        config.create_image_type = image_type

        # Add the new image layer.
        layer = self.viewer.add_image(data, rgb=True, name=unique_name)

        # TODO_OCTREE: We've not (yet?) added OctreeImage-specific
        # arguments to the OctreeImage constructor, because Layer class
        # arguments are special. They have to match the viewer add_image()
        # method?
        #
        # So for now we just set these values after construction, which is
        # kind of odd. And the class has to handle the value changing on
        # the fly. It would be better if they could be arguments or
        # passed in on construction somehow.
        layer.tile_size = image_settings.tile_size



import json
import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile

import numpy as np
import pytest

from .. import timers


@contextmanager
def temporary_file(suffix=''):
    """Yield a writeable temporary filename that is deleted on context exit.
    Parameters
    ----------
    suffix : string, optional
        The suffix for the file.
    """
    tempfile_stream = NamedTemporaryFile(suffix=suffix, delete=False)
    tempfile = tempfile_stream.name
    tempfile_stream.close()
    yield tempfile
    os.remove(tempfile)


def _trace_file_okay(trace_path: str) -> bool:
    """For now okay just means valid JSON and not empty."""


@pytest.mark.perfmon
def test_trace_on_start(make_test_viewer):
    """Make sure napari can write a perfmon trace file."""
    with temporary_file('json') as trace_path:
        timers.start_trace_file(trace_path)

        viewer = make_test_viewer()
        data = np.random.random((10, 15))
        viewer.add_image(data)
        viewer.close()

        timers.stop_trace_file()

        # Make sure file exists and is not empty.
        assert Path(trace_path).exists(), "Trace file not written"
        assert Path(trace_path).stat().st_size > 0, "Trace file is empty"

        # Assert every event contains every important field.
        with open(trace_path) as infile:
            data = json.load(infile)
            assert len(data) > 0
            for event in data:
                for field in ['pid', 'tid', 'name', 'ph', 'ts', 'args']:
                    assert field in event



"""ChunkKey and ChunkRequest classes.
"""
import contextlib
import logging
from typing import NamedTuple, Optional, Tuple

import numpy as np

from ....layers.base.base import Layer
from ....types import ArrayLike, Dict
from ....utils.perf import PerfEvent, block_timer
from ._utils import get_data_id

LOGGER = logging.getLogger("napari.async")

# We convert slices to tuple for hashing.
SliceTuple = Tuple[Optional[int], Optional[int], Optional[int]]


def _flatten(indices) -> tuple:
    """Return a flat tuple of integers to represent the indices.

    Slice objects are not hashable, so we convert them.
    """
    result = []
    for x in indices:
        if isinstance(x, slice):
            result.extend([x.start, x.stop, x.step])
        else:
            result.append(x)
    return tuple(result)


class LayerKey(NamedTuple):
    """The key for a layer and its important properties.

    Attributes
    ----------
    layer_id : int
        The id of the layer making the request.
    data_id : int
        The id of the data in the layer.
    data_level : int
        The level in the data (for multi-scale).
    indices : Tuple[Optional[slice], ...]
        The indices of the slice.
    """

    layer_id: int
    data_id: int
    data_level: int
    indices: Tuple[Optional[slice], ...]

    def _get_hash_values(self):
        return (
            self.layer_id,
            self.data_id,
            self.data_level,
            _flatten(self.indices),
        )


class ChunkKey:
    """The key for one single ChunkRequest.

    Parameters
    ----------
    layer : Layer
        The layer which is loading the chunk.
    indices : Indices
        The indices of the layer we are loading.

    Attributes
    ----------
    layer_key : LayerKey
        The layer specific parts of the key.
    key : Tuple
        The combined key, everything hashed together.
    """

    def __init__(self, layer: Layer, indices: Tuple[Optional[slice], ...]):
        self.layer_key = LayerKey(
            id(layer), get_data_id(layer), layer._data_level, indices
        )

        self.key = hash(self._get_hash_values())

    def _get_hash_values(self):
        return self.layer_key._get_hash_values()

    def __str__(self):
        layer_key = self.layer_key
        return (
            f"layer_id={layer_key.layer_id} data_id={layer_key.data_id} "
            f"data_level={layer_key.data_level} indices={layer_key.indices}"
        )

    def __eq__(self, other):
        return self.key == other.key


class ChunkRequest:
    """A request asking the ChunkLoader to load one or more arrays.

    Parameters
    ----------
    key : ChunkKey
        The key of the request.
    chunks : Dict[str, ArrayLike]
        The chunk arrays we need to load.

    Attributes
    ----------
    key : ChunkKey
        The key of the request.
    chunks : Dict[str, ArrayLike]
        The chunk arrays we need to load.
    timers : Dict[str, PerfEvent]
        Timing information about chunk load time.
    """

    def __init__(self, key: ChunkKey, chunks: Dict[str, ArrayLike]):
        # Make sure chunks dict is what we expect.
        for chunk_key, array in chunks.items():
            assert isinstance(chunk_key, str)
            assert array is not None

        self.key = key
        self.chunks = chunks

        self.timers: Dict[str, PerfEvent] = {}

    @property
    def data_id(self) -> int:
        return self.key.layer_key.data_id

    @property
    def num_chunks(self) -> int:
        """Return the number of chunks in this request."""
        return len(self.chunks)

    @property
    def num_bytes(self) -> int:
        """Return the number of bytes that were loaded."""
        return sum(array.nbytes for array in self.chunks.values())

    @property
    def in_memory(self) -> bool:
        """Return True if all chunks are ndarrays."""
        return all(isinstance(x, np.ndarray) for x in self.chunks.values())

    @contextlib.contextmanager
    def chunk_timer(self, name):
        """Time a block of code and save the PerfEvent in self.timers.

        We want to time our loads whether perfmon is enabled or not, since
        the auto-async feature needs to work in all cases.

        Parameters
        ----------
        name : str
            The name of the timer.

        Yields
        ------
        PerfEvent
            The timing event for the block.
        """
        with block_timer(name) as event:
            yield event
        self.timers[name] = event

    def load_chunks(self):
        """Load all of our chunks now in this thread.

        We time the overall load with the special name "load_chunks" and then
        we time each chunk as it loads, using it's array name as the key.
        """
        with self.chunk_timer("ChunkRequest.load_chunks"):
            for key, array in self.chunks.items():
                with self.chunk_timer(key):
                    loaded_array = np.asarray(array)
                    self.chunks[key] = loaded_array

    def transpose_chunks(self, order: tuple) -> None:
        """Transpose all our chunks.

        Parameters
        ----------
        order
            Transpose the chunks into this order.
        """
        for key, array in self.chunks.items():
            self.chunks[key] = array.transpose(order)

    @property
    def image(self) -> Optional[ArrayLike]:
        """The image chunk or None.

        Returns
        -------
        Optional[ArrayLike]
            The image chunk or None if we don't have one.
        """
        return self.chunks.get('image')

    @property
    def thumbnail_source(self):
        """The chunk to use as the thumbnail_source or None.

        Returns
        -------
        Optional[ArrayLike]
            The thumbnail_source chunk or None if we don't have one.
        """
        try:
            return self.chunks['thumbnail_source']
        except KeyError:
            # No thumbnail_source so return the image instead. For single-scale
            # we use the image as the thumbnail_source.
            return self.chunks.get('image')



"""Widgets for QtRender.

These were created for QtRender, but they are generic. Maybe we could switch
from these to something more napari-standard. These were just created
naively without knowing too much about napari's Qt code.
"""
from typing import Callable

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QSpinBox,
    QTableWidget,
    QTableWidgetItem,
    QWidget,
)


class QtLabeledSpinBox(QWidget):
    """A label plus a SpinBox.

    Parameters
    ----------
    label_text : str
        The label shown next to the spin box.
    initial_value : int
        The initial value of the spin box.
    spin_range : range
        The min/max/step of the spin box.
    connect : Callable[[int], None]
        Called when the user changes the value of the spin box.

    Attributes
    ----------
    spin : QSpinBox
        The spin box.
    """

    def __init__(
        self,
        label_text: str,
        initial_value: int,
        spin_range: range,
        connect: Callable[[int], None] = None,
    ):
        super().__init__()

        self._connect = connect
        layout = QHBoxLayout()

        self.spin = self._create(initial_value, spin_range)

        layout.addWidget(QLabel(label_text))
        layout.addWidget(self.spin)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def _create(self, initial_value: int, spin_range: range) -> QSpinBox:
        """Return the configured QSpinBox.

        Parameters
        ----------
        initial_value : int
            The initial value of the QSpinBox.
        spin_range : range
            The start/stop/step of the QSpinBox.

        Return
        ------
        QSpinBox
            The configured QSpinBox.
        """
        spin = QSpinBox()

        spin.setMinimum(spin_range.start)
        spin.setMaximum(spin_range.stop)
        spin.setSingleStep(spin_range.step)
        spin.setValue(initial_value)

        spin.setKeyboardTracking(False)  # correct?
        spin.setAlignment(Qt.AlignCenter)

        spin.valueChanged.connect(self._on_change)
        return spin

    def _on_change(self, value: int) -> None:
        """Called when the spin spin value was changed.

        value : int
            The new value of the SpinBox.
        """
        # We must clearFocus or it double-steps, no idea why.
        self.spin.clearFocus()

        if self._connect is not None:
            self._connect(value)  # Notify if we have a connection.


class QtLabeledComboBox(QWidget):
    """A generic ComboBox with a label.

    Provide an options dict. The keys will be displayed as the text options
    available to the user. The values are used by our set_value() and
    get_value() methods.

    Parameters
    ----------
    label : str
        The text label for the control.
    options : dict
        We display the keys and return the values.
    callback : Callable[[int], None]
        Called when the value is changed by the user.
    """

    def __init__(
        self, label: str, options: dict, callback: Callable[[int], None] = None
    ):
        super().__init__()
        self.options = options
        layout = QHBoxLayout()

        self.combo = QComboBox()
        self.combo.addItems(list(options.keys()))

        if callback is not None:
            self.combo.activated[int].connect(callback)

        layout.addWidget(QLabel(label))
        layout.addWidget(self.combo)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def set_value(self, value: str) -> None:
        """Make this value the selected item in the combo box.

        Parameters
        ----------
        value : str
            Set the combo box to this value.
        """
        for index, opt_value in enumerate(self.options.values()):
            if opt_value == value:
                self.combo.setCurrentIndex(index)

    def set_index(self, index: int) -> None:
        """Set what the combo box is showing based on the index.

        Parameters
        ----------
        index : int
            Set the combo box to this index.
        """
        self.combo.setCurrentIndex(index)

    def get_value(self) -> str:
        """Get the current value of the combo box.

        Return
        ------
        str
            The value currently selected in the combo box.
        """
        text = self.combo.currentText()
        return self.options[text]


class QtSimpleTable(QTableWidget):
    """A table of keys and values."""

    def __init__(self):
        super().__init__()
        self.verticalHeader().setVisible(False)
        self.horizontalHeader().setVisible(False)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.resizeRowsToContents()
        self.setShowGrid(False)

    def set_values(self, values: dict) -> None:
        """Populate the table with keys and values.

        values : dict
            Populate with these keys and values.
        """
        self.setRowCount(len(values))
        self.setColumnCount(2)
        for i, (key, value) in enumerate(values.items()):
            self.setItem(i, 0, QTableWidgetItem(key))
            self.setItem(i, 1, QTableWidgetItem(value))



from .event import EmitterGroup, Event, EventEmitter  # isort:skip
from .containers._evented_list import EventedList
from .containers._nested_list import NestableEventedList
from .containers._typed import TypedMutableSequence
from .dataclass import dataclass
from .types import SupportsEvents

__all__ = [
    'dataclass',
    'EmitterGroup',
    'Event',
    'EventedList',
    'EventEmitter',
    'NestableEventedList',
    'SupportsEvents',
    'TypedMutableSequence',
]



import sys
from contextlib import contextmanager
from os.path import dirname, join

from qtpy.QtCore import Qt
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QApplication, QSplashScreen

from ..utils.perf import perf_config
from .exceptions import ExceptionHandler


def _create_application(argv) -> QApplication:
    """Create our QApplication.

    Notes
    -----
    Substitute QApplicationWithTracing when using perfmon.

    With IPython/Jupyter we call convert_app_for_tracing() which deletes
    the QApplication and creates a new one. However here with gui_qt we
    need to create the correct QApplication up front, or we will crash.
    We'll crash because we'd be deleting the QApplication after we created
    QWidgets with it, such as we do for the splash screen.
    """
    if perf_config and perf_config.trace_qt_events:
        from .tracing.qt_event_tracing import QApplicationWithTracing

        return QApplicationWithTracing(argv)
    else:
        return QApplication(argv)


@contextmanager
def gui_qt(*, startup_logo=False, gui_exceptions=False):
    """Start a Qt event loop in which to run the application.

    Parameters
    ----------
    startup_logo : bool, optional
        Show a splash screen with the napari logo during startup.
    gui_exceptions : bool, optional
        Whether to show uncaught exceptions in the GUI, by default they will be
        shown in the console that launched the event loop.

    Notes
    -----
    This context manager is not needed if running napari within an interactive
    IPython session. In this case, use the ``%gui qt`` magic command, or start
    IPython with the Qt GUI event loop enabled by default by using
    ``ipython --gui=qt``.
    """
    splash_widget = None
    app = QApplication.instance()
    if not app:
        # automatically determine monitor DPI.
        # Note: this MUST be set before the QApplication is instantiated
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
        # if this is the first time the Qt app is being instantiated, we set
        # the name, so that we know whether to raise_ in Window.show()
        app = _create_application(sys.argv)
        app.setApplicationName('napari')
        if startup_logo:
            logopath = join(dirname(__file__), '..', 'resources', 'logo.png')
            pm = QPixmap(logopath).scaled(
                360, 360, Qt.KeepAspectRatio, Qt.SmoothTransformation
            )
            splash_widget = QSplashScreen(pm)
            splash_widget.show()
            app._splash_widget = splash_widget
    else:
        app._existed = True

    # instantiate the exception handler
    exception_handler = ExceptionHandler(gui_exceptions=gui_exceptions)
    sys.excepthook = exception_handler.handle

    try:
        yield app
    except Exception:
        exception_handler.handle(*sys.exc_info())

    # if the application already existed before this function was called,
    # there's no need to start it again.  By avoiding unnecessary calls to
    # ``app.exec_``, we avoid blocking.
    if app.applicationName() == 'napari':
        if splash_widget and startup_logo:
            splash_widget.close()
        app.exec_()



import pytest

from napari.components import Dims


def test_ndim():
    """
    Test number of dimensions including after adding and removing dimensions.
    """
    dims = Dims()
    assert dims.ndim == 2

    dims = Dims(4)
    assert dims.ndim == 4

    dims = Dims(2)
    assert dims.ndim == 2

    dims.ndim = 10
    assert dims.ndim == 10

    dims.ndim = 5
    assert dims.ndim == 5


def test_display():
    """
    Test display setting.
    """
    dims = Dims(4)
    assert dims.order == [0, 1, 2, 3]
    assert dims.ndisplay == 2

    dims.order = [2, 3, 1, 0]
    assert dims.order == [2, 3, 1, 0]


def test_order_with_init():
    dims = Dims(3, order=[0, 2, 1])
    assert dims.order == [0, 2, 1]


def test_labels_with_init():
    dims = Dims(3, axis_labels=['x', 'y', 'z'])
    assert dims.axis_labels == ['x', 'y', 'z']


def test_wrong_order():
    with pytest.raises(ValueError):
        Dims(3, order=range(2))


def test_wrong_labels():
    with pytest.raises(ValueError):
        Dims(3, axis_labels=['a', 'b'])


def test_keyword_only_dims():
    with pytest.raises(TypeError):
        Dims(3, [1, 2, 3])


def test_point():
    """
    Test point setting.
    """
    dims = Dims(4)
    assert dims.point == [0] * 4

    dims.set_range(3, (0, 5, 1))
    dims.set_point(3, 4)
    assert dims.point == [0, 0, 0, 4]

    dims.set_range(2, (0, 5, 1))
    dims.set_point(2, 1)
    assert dims.point == [0, 0, 1, 4]


def test_range():
    """
    Tests range setting.
    """
    dims = Dims(4)
    assert dims.range == [(0, 2, 1)] * 4

    dims.set_range(3, (0, 4, 2))
    assert dims.range == [(0, 2, 1)] * 3 + [(0, 4, 2)]


def test_axis_labels():
    dims = Dims(4)
    assert dims.axis_labels == ['0', '1', '2', '3']


def test_order_when_changing_ndim():
    """
    Test order of the dims when changing the number of dimensions.
    """
    dims = Dims(4)
    dims.set_range(0, (0, 4, 1))
    dims.set_point(0, 2)

    dims.ndim = 5
    # Test that new dims get appended to the beginning of lists
    assert dims.point == [0, 2, 0, 0, 0]
    assert dims.order == [0, 1, 2, 3, 4]
    assert dims.axis_labels == ['0', '1', '2', '3', '4']

    dims.set_range(2, (0, 4, 1))
    dims.set_point(2, 3)
    dims.ndim = 3
    # Test that dims get removed from the beginning of lists
    assert dims.point == [3, 0, 0]
    assert dims.order == [0, 1, 2]
    assert dims.axis_labels == ['2', '3', '4']


def test_labels_order_when_changing_dims():
    dims = Dims(4)
    dims.ndim = 5
    assert dims.axis_labels == ['0', '1', '2', '3', '4']


@pytest.mark.parametrize(
    "ndim, ax_input, expected", [(2, 1, 1), (2, -1, 1), (4, -3, 1)]
)
def test_assert_axis_in_bounds(ndim, ax_input, expected):
    dims = Dims(ndim)
    actual = dims._assert_axis_in_bounds(ax_input)
    assert actual == expected


@pytest.mark.parametrize("ndim, ax_input", [(2, 2), (2, -3)])
def test_assert_axis_out_of_bounds(ndim, ax_input):
    dims = Dims(ndim)
    with pytest.raises(ValueError):
        dims._assert_axis_in_bounds(ax_input)


def test_axis_labels_str_to_list():
    dims = Dims()
    dims.axis_labels = 'TX'
    assert dims.axis_labels == ['T', 'X']


def test_roll():
    """Test basic roll behavior."""
    dims = Dims(ndim=4)
    dims.set_range(0, (0, 10, 1))
    dims.set_range(1, (0, 10, 1))
    dims.set_range(2, (0, 10, 1))
    dims.set_range(3, (0, 10, 1))
    assert dims.order == [0, 1, 2, 3]
    dims._roll()
    assert dims.order == [3, 0, 1, 2]
    dims._roll()
    assert dims.order == [2, 3, 0, 1]


def test_roll_skip_dummy_axis_1():
    """Test basic roll skips axis with length 1."""
    dims = Dims(ndim=4)
    dims.set_range(0, (0, 0, 1))
    dims.set_range(1, (0, 10, 1))
    dims.set_range(2, (0, 10, 1))
    dims.set_range(3, (0, 10, 1))
    assert dims.order == [0, 1, 2, 3]
    dims._roll()
    assert dims.order == [0, 3, 1, 2]
    dims._roll()
    assert dims.order == [0, 2, 3, 1]


def test_roll_skip_dummy_axis_2():
    """Test basic roll skips axis with length 1 when not first."""
    dims = Dims(ndim=4)
    dims.set_range(0, (0, 10, 1))
    dims.set_range(1, (0, 0, 1))
    dims.set_range(2, (0, 10, 1))
    dims.set_range(3, (0, 10, 1))
    assert dims.order == [0, 1, 2, 3]
    dims._roll()
    assert dims.order == [3, 1, 0, 2]
    dims._roll()
    assert dims.order == [2, 1, 3, 0]


def test_roll_skip_dummy_axis_3():
    """Test basic roll skips all axes with length 1."""
    dims = Dims(ndim=4)
    dims.set_range(0, (0, 10, 1))
    dims.set_range(1, (0, 0, 1))
    dims.set_range(2, (0, 10, 1))
    dims.set_range(3, (0, 0, 1))
    assert dims.order == [0, 1, 2, 3]
    dims._roll()
    assert dims.order == [2, 1, 0, 3]
    dims._roll()
    assert dims.order == [0, 1, 2, 3]


def test_changing_focus(qtbot):
    """Test changing focus updates the last_used prop."""
    # too-few dims, should have no sliders to update
    dims = Dims(2)
    assert dims.last_used is None
    dims._focus_down()
    dims._focus_up()
    assert dims.last_used is None

    dims.ndim = 5
    # Note that with no view attached last used remains
    # None even though new non-displayed dimensions added
    assert dims.last_used is None
    dims._focus_down()
    assert dims.last_used == 2
    dims._focus_down()
    assert dims.last_used == 1
    dims._focus_up()
    assert dims.last_used == 2
    dims._focus_up()
    assert dims.last_used == 0
    dims._focus_down()
    assert dims.last_used == 2



"""Create test images

This is a throw-away file for creating a test image for octree rendering
development. If we keep the in-napari creation of test images long term,
we'll probably want a better system for creating them.

Also, long term we probably do not want to use PIL at all, since it's
not a desirable dependency.
"""
from typing import Tuple

import numpy as np
from PIL import Image, ImageDraw, ImageFont

from ....layers.image.experimental import (
    SliceConfig,
    create_multi_scale_from_image,
)
from ....utils.perf import block_timer


class TextWriter:
    """Write text into a PIL image."""

    def __init__(self, draw: ImageDraw.Draw, font_str: str, color: str):
        self.draw = draw
        try:
            self.font = ImageFont.truetype(font_str, size=74)
        except OSError:
            self.font = ImageFont.load_default()
        self.color = color

    def get_text_size(self, text: str) -> Tuple[int, int]:
        """Get size of the given text."""
        return self.font.getsize(text)

    def write_text(self, pos: Tuple[int, int], text: str) -> None:
        """Write the text into the image."""
        self.draw.text(pos, text, fill=self.color, font=self.font)


def draw_text_grid(image, text: str) -> None:
    """Draw some text into the given image in a grid.

    Parameters
    ----------
    test : str
        The text to draw. For example a slice index like "3".
    """
    color = 'rgb(255, 255, 255)'  # white
    width, height = image.size
    draw = ImageDraw.Draw(image)

    writer = TextWriter(draw, 'Arial Black.ttf', color)
    (text_width, text_height) = writer.get_text_size(text)

    text_size = 100  # hack, approx guess with some padding

    rows, cols = int(height / text_size), int(width / text_size)

    for row in range(rows + 1):
        for col in range(cols + 1):
            pos = [
                (col / cols) * image.width - text_width / 2,  # x
                (row / rows) * image.height - text_height / 2,  # y
            ]

            writer.write_text(pos, text)

    draw.rectangle([0, 0, image.width, image.height], outline=color, width=5)


def create_test_image(text, slice_config: SliceConfig) -> np.ndarray:
    """Create a test image for testing tiled rendering.

    The test image just has digits all over it. The digits will typically
    be used to show the slice number like "0" or "42".

    slice_config: SliceConfig
        The base image shape and other details.
    """
    text = str(text)  # Might be an int.

    # Image.new wants (width, height) so swap them.
    image_size = slice_config.image_shape[::-1]

    # Create the image, draw on the text, return it.
    image = Image.new('RGB', image_size)
    draw_text_grid(image, text)
    return np.array(image)


def create_test_image_multi(text, slice_config: SliceConfig) -> np.ndarray:
    """Create a multiscale test image for testing tiled rendering.

    The test image is blank with digits all over it. The digits will
    typically be used to show the slice number.

    slice_config: SliceConfig
        The base image shape and other details.
    """
    text = str(text)  # Convert in case it's an int.

    # Image.new wants (width, height) so swap them.
    image_size = slice_config.base_shape[::-1]

    # Create the image, draw on the text, return it.
    image = Image.new('RGB', image_size)
    draw_text_grid(image, text)
    data = np.array(image)

    with block_timer("create_multi_scale_from_image", print_time=True):
        # This is create all the octree layers, each coarser one
        # downsampled from the previous level.
        return create_multi_scale_from_image(data, slice_config.tile_size)



"""Layers are the viewable objects that can be added to a viewer.

Custom layers must inherit from Layer and pass along the
`visual node <http://vispy.org/scene.html#module-vispy.scene.visuals>`_
to the super constructor.
"""

from ..utils.misc import all_subclasses
from .base import Layer
from .image import Image
from .labels import Labels
from .points import Points
from .shapes import Shapes
from .surface import Surface
from .vectors import Vectors
from .tracks import Tracks

NAMES = {subclass.__name__.lower() for subclass in all_subclasses(Layer)}
del all_subclasses



"""Napari Configuration.
"""
import os


def _set(env_var: str) -> bool:
    """Return True if the env variable is set and non-zero.

    Return
    ------
    bool
        True if the env var was set to a non-zero value.
    """
    return os.getenv(env_var, "0") != "0"


# Added this temporarily for octree debugging. The welcome visual causes
# breakpoints to hit in image visual code. It's easier if we don't show it.
allow_welcome_visual = True


"""
Experimental Image Configuration.

Async loading and octree image rendering are two separate but related
experimental features.

Async Loading
-------------
Image layers will use the ChunkLoader to load data instead of loading
the data directly. Image layers will not call np.asarray() in the GUI
thread. The ChunkLoader will call np.asarray() in a worker thread. That
means any IO or computation done as part of the load will not block the
GUI thread.

Set NAPARI_ASYNC=1 to turn on async loading with default settings.

Set NAPARI_ASYNC=/tmp/async.json to enable async loading with
a configuration file. See config settings here:
    napari.components.experimental.chunk._config.py

Octree Rendering
----------------
Image layers use an octree for rendering. The octree organizes the image
into chunks/tiles. Only a subset of those chunks/tiles are loaded and
drawn at a time. Octree rendering is very WIP and not ready yet.

Set NAPARI_OCTREE=1 to enable experimental octree visuals. This
will also turn on async loading, however to configure async
loading you can set NAPARI_ASYNC to a config file path as above.

Future
------
When we're done with octree development we want to return to one single
image class, with one single type of visual. All of this config complexity
is temporary.
"""

async_octree = _set("NAPARI_OCTREE")
async_loading = _set("NAPARI_ASYNC") or async_octree

"""
Image Layer Creation

With octree enabled the QtTestImage widget allows creating 3 types of image
layers/visuals. This lets us compare each implemention.
"""

# Image layer, the original one.
CREATE_IMAGE_NORMAL = 1

# OctreeImage layer with the VispyCompoundImageLayer visual.
# This visual uses a separate ImageVisual for each tile.
CREATE_IMAGE_COMPOUND = 2

# OctreeImage layer with the VispyTiledImageLayer visual.
# This is a single visual that can draw many tiles.
CREATE_IMAGE_TILED = 3

# The default type of image the QtTestImage widget will create. If the user
# selects another type of image, the widget will change this config value.
create_image_type = CREATE_IMAGE_TILED


def create_octree_image() -> bool:
    """Return True if we should create an OctreeImage layer.

    Only create an OctreeImage layer if octree is enabled AND we are
    creating an octree-based image type.

    Return
    ------
    bool
        True if we should create an OctreeImage layer.
    """
    return async_octree and (create_image_type != CREATE_IMAGE_NORMAL)



import numpy as np

from ..utils.colormaps.standardize_color import transform_color
from ..utils.events import EmitterGroup
from ._viewer_constants import Position


class ScaleBar:
    """Scale bar indicating size in world coordinates.

    Attributes
    ----------
    background_color : np.ndarray
        Background color of canvas. If scale bar is not colored
        then it has the color opposite of this color.
    colored : bool
        If scale bar are colored or not. If colored then
        default color is magenta. If not colored than
        scale bar color is the opposite of the canvas
        background.
    events : EmitterGroup
        Event emitter group
    position : str
        Position of the scale bar in the canvas. Must be one of
        'top left', 'top right', 'bottom right', 'bottom left'.
        Default value is 'bottom right'.
    ticks : bool
        If scale bar has ticks at ends or not.
    visible : bool
        If scale bar is visible or not.
    """

    def __init__(self):

        # Events:
        self.events = EmitterGroup(
            source=self,
            auto_connect=True,
            visible=None,
            colored=None,
            position=None,
            ticks=None,
        )
        self._visible = False
        self._colored = False
        self._background_color = np.array([1, 1, 1])
        self._ticks = True
        self._position = Position('bottom_right')

    @property
    def visible(self):
        """bool: If scale bar is visible or not."""
        return self._visible

    @visible.setter
    def visible(self, visible):
        self._visible = visible
        self.events.visible()

    @property
    def colored(self):
        """bool: If scale bar is colored or not."""
        return self._colored

    @colored.setter
    def colored(self, colored):
        self._colored = colored
        self.events.colored()

    @property
    def background_color(self):
        """np.ndarray: RGBA color."""
        return self._background_color

    @background_color.setter
    def background_color(self, background_color):
        self._background_color = transform_color(background_color)[0]
        self.events.colored()

    @property
    def ticks(self):
        """bool: If scale bar has ticks or not."""
        return self._ticks

    @ticks.setter
    def ticks(self, ticks):
        self._ticks = ticks
        self.events.ticks()

    @property
    def position(self):
        """Position: Location of scale bar in canvas.

        Postion.TOP_LEFT
             Top left of the canvas.
        Postion.TOP_RIGHT
             Top right of the canvas.
        Postion.BOTTOM_RIGHT
             Bottom right of the canvas.
        Postion.BOTTOM_LEFT
             Bottom left of the canvas.
        """
        return str(self._position)

    @position.setter
    def position(self, position):
        self._position = Position(position)
        self.events.position()



"""OpenGL Utilities.
"""
from contextlib import contextmanager
from functools import lru_cache
from typing import Tuple

from vispy.app import Canvas
from vispy.gloo import gl
from vispy.gloo.context import get_current_canvas


@contextmanager
def _opengl_context():
    """Assure we are running with a valid OpenGL context.

    Only create a Canvas is one doesn't exist. Creating and closing a
    Canvas causes vispy to process Qt events which can cause problems.
    Ideally call opengl_context() on start after creating your first
    Canvas. However it will work either way.
    """
    canvas = Canvas(show=False) if get_current_canvas() is None else None
    try:
        yield
    finally:
        if canvas is not None:
            canvas.close()


@lru_cache()
def get_max_texture_sizes() -> Tuple[int, int]:
    """Return the maximum texture sizes for 2D and 3D rendering.

    If this function is called without an OpenGL context it will create a
    temporary non-visible Canvas. Either way the lru_cache means subsequent
    calls to thing function will return the original values without
    actually running again.

    Returns
    -------
    Tuple[int, int]
        The max textures sizes for (2d, 3d) rendering.
    """
    with _opengl_context():
        max_size_2d = gl.glGetParameter(gl.GL_MAX_TEXTURE_SIZE)

    if max_size_2d == ():
        max_size_2d = None

    # vispy doesn't expose GL_MAX_3D_TEXTURE_SIZE so hard coding for now.
    # MAX_TEXTURE_SIZE_3D = gl.glGetParameter(gl.GL_MAX_3D_TEXTURE_SIZE)
    # if MAX_TEXTURE_SIZE_3D == ():
    #    MAX_TEXTURE_SIZE_3D = None
    max_size_3d = 2048

    return max_size_2d, max_size_3d



from abc import ABC, abstractmethod

import numpy as np
from vispy.visuals.transforms import MatrixTransform

from .utils_gl import get_max_texture_sizes


class VispyBaseLayer(ABC):
    """Base object for individual layer views

    Meant to be subclassed.

    Parameters
    ----------
    layer : napari.layers.Layer
        Layer model.
    node : vispy.scene.VisualNode
        Central node with which to interact with the visual.

    Attributes
    ----------
    layer : napari.layers.Layer
        Layer model.
    node : vispy.scene.VisualNode
        Central node with which to interact with the visual.
    scale : sequence of float
        Scale factors for the layer visual in the scenecanvas.
    translate : sequence of float
        Translation values for the layer visual in the scenecanvas.
    MAX_TEXTURE_SIZE_2D : int
        Max texture size allowed by the vispy canvas during 2D rendering.
    MAX_TEXTURE_SIZE_3D : int
        Max texture size allowed by the vispy canvas during 2D rendering.


    Extended Summary
    ----------------
    _master_transform : vispy.visuals.transforms.MatrixTransform
        Transform positioning the layer visual inside the scenecanvas.
    """

    def __init__(self, layer, node):
        super().__init__()

        self.layer = layer
        self._array_like = False
        self.node = node

        (
            self.MAX_TEXTURE_SIZE_2D,
            self.MAX_TEXTURE_SIZE_3D,
        ) = get_max_texture_sizes()

        self.layer.events.refresh.connect(lambda e: self.node.update())
        self.layer.events.set_data.connect(self._on_data_change)
        self.layer.events.visible.connect(self._on_visible_change)
        self.layer.events.opacity.connect(self._on_opacity_change)
        self.layer.events.blending.connect(self._on_blending_change)
        self.layer.events.scale.connect(self._on_matrix_change)
        self.layer.events.translate.connect(self._on_matrix_change)
        self.layer.events.rotate.connect(self._on_matrix_change)
        self.layer.events.shear.connect(self._on_matrix_change)
        self.layer.events.affine.connect(self._on_matrix_change)

    @property
    def _master_transform(self):
        """vispy.visuals.transforms.MatrixTransform:
        Central node's firstmost transform.
        """
        # whenever a new parent is set, the transform is reset
        # to a NullTransform so we reset it here
        if not isinstance(self.node.transform, MatrixTransform):
            self.node.transform = MatrixTransform()

        return self.node.transform

    @property
    def translate(self):
        """sequence of float: Translation values."""
        return self._master_transform.matrix[-1, :]

    @property
    def scale(self):
        """sequence of float: Scale factors."""
        matrix = self._master_transform.matrix[:-1, :-1]
        _, upper_tri = np.linalg.qr(matrix)
        return np.diag(upper_tri).copy()

    @property
    def order(self):
        """int: Order in which the visual is drawn in the scenegraph.

        Lower values are closer to the viewer.
        """
        return self.node.order

    @order.setter
    def order(self, order):
        self.node.order = order

    @abstractmethod
    def _on_data_change(self, event=None):
        raise NotImplementedError()

    def _on_visible_change(self, event=None):
        self.node.visible = self.layer.visible

    def _on_opacity_change(self, event=None):
        self.node.opacity = self.layer.opacity

    def _on_blending_change(self, event=None):
        self.node.set_gl_state(self.layer.blending)
        self.node.update()

    def _on_matrix_change(self, event=None):
        transform = self.layer._transforms.simplified.set_slice(
            self.layer._dims.displayed
        )
        # convert NumPy axis ordering to VisPy axis ordering
        # by reversing the axes order and flipping the linear
        # matrix
        translate = transform.translate[::-1]
        matrix = transform.linear_matrix[::-1, ::-1].T

        # Embed in the top left corner of a 4x4 affine matrix
        affine_matrix = np.eye(4)
        affine_matrix[: matrix.shape[0], : matrix.shape[1]] = matrix
        affine_matrix[-1, : len(translate)] = translate

        if self._array_like and self.layer._dims.ndisplay == 2:
            # Perform pixel offset to shift origin from top left corner
            # of pixel to center of pixel.
            # Note this offset is only required for array like data in
            # 2D.
            offset_matrix = (
                self.layer._transforms['data2world']
                .set_slice(self.layer._dims.displayed)
                .linear_matrix
            )
            offset = -offset_matrix @ np.ones(offset_matrix.shape[1]) / 2
            # Convert NumPy axis ordering to VisPy axis ordering
            # and embed in full affine matrix
            affine_offset = np.eye(4)
            affine_offset[-1, : len(offset)] = offset[::-1]
            affine_matrix = affine_matrix @ affine_offset
        self._master_transform.matrix = affine_matrix

    def _reset_base(self):
        self._on_visible_change()
        self._on_opacity_change()
        self._on_blending_change()
        self._on_matrix_change()

    def _on_camera_move(self, event=None):
        """Camera was moved."""
        pass



"""QtMiniMap widget.

Creates a bitmap that shows which octree tiles are being viewed.
"""
import math
from typing import NamedTuple

import numpy as np
from qtpy.QtGui import QImage, QPixmap
from qtpy.QtWidgets import QLabel

from ....layers.image.experimental import OctreeIntersection, OctreeLevel
from ....layers.image.experimental.octree_image import OctreeImage

# Longest edge of map bitmap in pixels. So if in a odd shape it does not
# become bigger than this in either direction.
MAP_SIZE = 220

# Seen tiles are inside the view.
COLOR_SEEN = (255, 0, 0, 255)  # red
COLOR_UNSEEN = (80, 80, 80, 255)  # gray

# The view bounds are draw on top of the seen/unseen tiles.
COLOR_VIEW = (227, 220, 111, 255)  # yellow

# Create a gap between the tiles.
TILE_EDGE = 1


class Rect(NamedTuple):
    """Rectangle that we can draw into a numpy array."""

    x: float
    y: float
    width: float
    height: float

    def draw(self, data: np.ndarray, color: np.ndarray) -> None:
        """Draw the rectangle into the array."""
        y0, y1 = int(self.y), int(self.y + self.height)
        x0, x1 = int(self.x), int(self.x + self.width)

        # Create a small border between tiles.
        y0 += TILE_EDGE
        y1 -= TILE_EDGE
        x0 += TILE_EDGE
        x1 -= TILE_EDGE

        data[y0:y1, x0:x1, :] = color


def _draw_view(data, intersection: OctreeIntersection) -> None:
    """Draw the view rectangle onto the map data.

    Parameters
    ----------
    data : np.ndarray
        Draw the view into this data.
    intersection : OctreeIntersection
        Draw the view in this intersection.
    """
    # Max (row, col) dimensions of the bitmap we are writing into.
    max_dim = np.array([data.shape[0] - 1, data.shape[1] - 1])

    # Convert normalized ranges into bitmap pixel ranges
    ranges = (intersection.normalized_range * max_dim).astype(int)

    # Write the view color into this rectangular regions.
    # TODO_OCTREE: must be a nicer way to index this?
    rows, cols = ranges[0], ranges[1]
    data[rows[0] : rows[1], cols[0] : cols[1], :] = COLOR_VIEW


def _draw_tiles(
    bitmap: np.ndarray, intersection: OctreeIntersection, scale: np.ndarray
) -> None:
    """Draw all the tiles, marking which are seen by the intersection.

    Parameters
    ----------
    intersection : OctreeIntersection
        The intersection we are drawing.
    scale_xy : Tuple[float, float]
        The scale to draw things that.
    """
    level = intersection.level
    shape = level.info.slice_config.base_shape

    y = 0
    for row in range(shape[0]):
        x = 0
        for col in range(shape[1]):
            octree_chunk = level.get_chunk(row, col)

            if octree_chunk is None:
                print(f"NO CHUNK: {row}, {col}")
                scaled = [0, 0]
                continue  # should not happen?

            scaled = octree_chunk.data.shape[:2] * scale
            rect = Rect(x, y, scaled[1], scaled[0])
            row, col = octree_chunk.location.row, octree_chunk.location.col

            color = (
                COLOR_SEEN
                if intersection.is_visible(row, col)
                else COLOR_UNSEEN
            )

            rect.draw(bitmap, color)
            x += scaled[1]
        y += scaled[0]


def _get_bitmap_shape(aspect: float) -> np.ndarray:
    """Get shape for the map bitmap.

    Parameters
    ----------
    aspect : float
        The width:height aspect ratio of the base image for the map.

    Return
    ------
    Tuple[int, int]
        The shape for the map bitmap.
    """
    depth = 4  # RGBA

    # Limit to at most MAP_SIZE pixels, in whichever dimension is the
    # bigger one. So it's not too huge even if an odd shape.
    if aspect > 1:
        # Limit the width, it is longer.
        return np.array((math.ceil(MAP_SIZE / aspect), MAP_SIZE, depth))

    # Limit the height, it is longer.
    return np.array((MAP_SIZE, math.ceil(MAP_SIZE * aspect), depth))


def _draw_intersection(intersection: OctreeIntersection) -> np.ndarray:
    """Return a bitmap that shows the tiles and the intersection.

    Parameters
    ----------
    intersection : OctreeIntersection
        Draw this intersection.

    Returns
    -------
    np.ndarray
        The bitmap showing the intersection.
    """
    aspect = intersection.level.info.slice_config.aspect_ratio
    level: OctreeLevel = intersection.level

    # Map shape plus RGBA depth.
    bitmap_shape = _get_bitmap_shape(aspect)
    data = np.zeros(bitmap_shape, dtype=np.uint8)

    # Scale the intersection down to fit in the bitmap.
    scale = bitmap_shape[:2] / level.info.image_shape

    # Draw all the tiles, the seen ones are drawn in red.
    _draw_tiles(data, intersection, scale)

    # Draw the view frustum in yellow.
    _draw_view(data, intersection)

    return data


class QtMiniMap(QLabel):
    """A small bitmap that shows the view bounds and which tiles are seen.

    Only works with OctreeImage layers.

    TODO_OCTREE: Consider this Qt bitmap rendering code just a rough
    proof-of-concept prototype. A real minimap should probably be a little
    OpenGL window, not a Qt bitmap. If we used OpenGL the GPU does most of
    the work and we get better quality and effects, like alpha and
    anti-aliasing. We do not want to spend a lot of time becoming good at
    "rendering" into Qt Bitmaps.

    Parameters
    ----------
    layer : OctreeImage
        The octree image we are viewing.
    """

    def __init__(self, layer: OctreeImage):
        super().__init__()
        self.layer = layer

    def update(self) -> None:
        """Update the minimap to show latest intersection."""
        # This actually performs the intersection, but it's very fast.
        intersection = self.layer.get_intersection()

        if intersection is not None:
            self._draw(intersection)

    def _draw(self, intersection: OctreeIntersection) -> None:
        """Draw the minimap showing the latest intersection.

        Update us with the new bitmap, we are a QLabel.

        Parameters
        ----------
        intersection : OctreeIntersection
            The intersection we are drawing on the mini map.
        """
        data = _draw_intersection(intersection)
        height, width = data.shape[:2]
        image = QImage(data, width, height, QImage.Format_RGBA8888)
        self.setPixmap(QPixmap.fromImage(image))



"""Octree utility classes.
"""
from typing import NamedTuple, Tuple

import numpy as np


class TestImageSettings(NamedTuple):
    """Settings for a test image we are creating."""

    base_shape: Tuple[int, int]
    tile_size: int


class NormalNoise(NamedTuple):
    """Noise with a normal distribution."""

    mean: float = 0
    std_dev: float = 0

    @property
    def is_zero(self) -> bool:
        """Return True if there is no noise at all.

        Return
        ------
        bool
            True if there is no noise at all.
        """
        return self.mean == 0 and self.std_dev == 0

    @property
    def get_value(self) -> float:
        """Get a random value.

        Return
        ------
        float
            The random value.
        """
        return np.random.normal(self.mean, self.std_dev)


class SliceConfig(NamedTuple):
    """Configuration for a tiled image.

    Attributes
    ----------
    base_shape : np.ndarray
        The base [height, width] shape of the entire full resolution image.
    num_levels : int
        The number of octree levels in the image.
    tile_size : int
        The default tile size. However each OctreeLevel has its own tile size
        which can override this.

    Notes
    -----
    This SliceConfig.tile_size will be used by the OctreeLevels in the tree
    in general. But the highest level OctreeLevel might use a larger size
    so that it can consist of a single chunk.

    For example we might be using 256x256 tiles in general. For best
    performance it might make sense to have octree levels such that the
    highest level fits inside a single 256x256 tiles.

    But if we are displaying user provided data, they did not know our tile
    size. Instead their root level might be something pretty big, like
    6000x6000. In that case we use 6000x6000 as the tile size in our root,
    so the root level consists of a single tile.

    TODO_OCTREE: we don't actually support larger size tiles yet! However
    it's still a good idea to assume that each OctreeLevel could have its
    own tile size.
    """

    base_shape: np.ndarray
    num_levels: int
    tile_size: int
    delay_ms: NormalNoise = NormalNoise()

    @property
    def aspect_ratio(self):
        """Return the width:height aspect ratio of the base image.

        For example HDTV resolution is 16:9 which is 1.77.
        """
        return self.base_shape[1] / self.base_shape[0]



"""QtRenderContainer class.
"""
from qtpy.QtWidgets import QStackedWidget

from ....components.viewer_model import ViewerModel
from .qt_render import QtRender


class QtRenderContainer(QStackedWidget):
    """Container widget for QtRender widgets.

    QtRender is a debug/developer widget for rendering and octree related
    functionality. We put up a QtRender for any later, but most of the
    controls are only visible for OctreeImage layers.

    Parameters
    ----------
    viewer : napari.components.ViewerModel
        Napari viewer containing the rendered scene, layers, and controls.

    Attributes
    ----------
    default_widget : QtRender
        A minimal version fo QtRender if no layer is selected.
    viewer : ViewerModel
        Napari viewer.
    widgets : dict
        Maps layer to its QtRender widget.
    """

    def __init__(self, viewer: ViewerModel):

        super().__init__()
        self.setProperty("emphasized", True)
        self.viewer = viewer

        self.setMouseTracking(True)
        self.setMinimumWidth(250)

        # We show QtRender even if no layer is selected. But in that case the
        # only control are to create test images.
        self.default_widget = QtRender(viewer)

        self._widgets = {}
        self.addWidget(self.default_widget)
        self._display(None)

        self.viewer.layers.events.inserted.connect(self._add)
        self.viewer.layers.events.removed.connect(self._remove)
        self.viewer.events.active_layer.connect(self._display)

    def _display(self, event):
        """Change the displayed controls to be those of the target layer.

        Parameters
        ----------
        event : Event
            Event with the target layer at `event.value`.
        """
        if event is None:
            layer = None
        else:
            layer = event.item

        if layer is None:
            self.setCurrentWidget(self.default_widget)
        else:
            controls = self._widgets[layer]
            self.setCurrentWidget(controls)

    def _add(self, event):
        """Add the controls target layer to the list of control widgets.

        Parameters
        ----------
        event : Event
            Event with the target layer at `event.value`.
        """
        layer = event.value
        controls = QtRender(self.viewer, layer)
        self.addWidget(controls)
        self._widgets[layer] = controls

    def _remove(self, event):
        """Remove the controls target layer from the list of control widgets.

        Parameters
        ----------
        event : Event
            Event with the target layer at `event.value`.
        """
        layer = event.value
        controls = self._widgets[layer]
        self.removeWidget(controls)
        controls.deleteLater()
        controls = None
        del self._widgets[layer]



"""QtOctreeInfo and QtOctreeInfoLayout classes.

Shows octree-specific information in the QtRender widget.
"""
import numpy as np
from qtpy.QtWidgets import QCheckBox, QFrame, QVBoxLayout

from ....components.experimental import chunk_loader
from ....layers.image.experimental.octree_image import NormalNoise, OctreeImage
from .qt_render_widgets import QtLabeledComboBox, QtSimpleTable


def _get_table_values(layer: OctreeImage) -> dict:
    """Get keys/values about this octree image for the table.

    layer : OctreeImage
        Get values for this octree image.
    """

    def _shape(shape) -> str:
        return f"{shape[1]}x{shape[0]}"

    level_info = layer.octree_level_info

    if level_info is None:
        return {}

    shape_in_tiles = level_info.shape_in_tiles
    num_tiles = shape_in_tiles[0] * shape_in_tiles[1]

    return {
        "Level": f"{layer.octree_level}",
        "Tiles": f"{_shape(shape_in_tiles)} = {num_tiles}",
        "Tile Shape": _shape([layer.tile_size, layer.tile_size]),
        "Layer Shape": _shape(level_info.image_shape),
    }


def _create_mean_combo(on_set) -> QtLabeledComboBox:
    """Return combo box for the mean noise setting.

    Return
    ------
    QtLabeledComboBox
        The new combo box.
    """
    options = {str(x): x for x in [0, 10, 20, 50, 100, 200, 500, 1000]}
    return QtLabeledComboBox("Mean Delay (ms)", options, on_set)


def _create_std_dev_combo(on_set) -> QtLabeledComboBox:
    """Return combo box for the std dev noise setting.

    Return
    ------
    QtLabeledComboBox
        The new combo box.
    """
    options = {
        str(x): x for x in [0, 1, 2, 5, 10, 20, 50, 100, 200, 500, 1000]
    }
    return QtLabeledComboBox("Std Dev (ms)", options, on_set)


class QtOctreeInfoLayout(QVBoxLayout):
    """OctreeImage specific information.

    Combo base to choose octree layer or set to AUTO for the normal rendering
    mode where the correct level is chosen automatically. (not working yet)

    Parameters
    ----------
    layer : OctreeImage
        Show octree info for this layer.
    on_set_level : Callable[[int], None]
        Call this when the octree level is changed.
    """

    def __init__(self, layer: OctreeImage):
        super().__init__()

        self.layer = layer

        def on_set_cache(value: int) -> None:
            chunk_loader.cache.enabled = value != 0

        def on_set_grid(value: int) -> None:
            self.layer.show_grid = value != 0

        def on_set_level(value: int) -> None:
            if value == 0:  # This is AUTO
                self.layer.freeze_level = False
            else:
                level = value - 1  # Account for AUTO at 0
                self.layer.freeze_level = True
                self.layer.octree_level = level

        def on_set_delay(_value: int) -> None:
            mean = self.mean_combo.get_value()
            std_dev = self.std_dev_combo.get_value()
            self.layer.delay_ms = NormalNoise(mean, std_dev)

        def on_set_track(value: int):
            self.layer.track_view = value != 0

        # Toggle the ChunkCache.
        cache_enabled = chunk_loader.cache.enabled
        self._create_checkbox("Chunk Cache", cache_enabled, on_set_cache)

        # Toggle tracking: drawn tiles track view as it moves.
        self._create_checkbox("Track View", layer.track_view, on_set_track)

        # Toggle debug grid drawn around tiles.
        self._create_checkbox("Show Grid", layer.show_grid, on_set_grid)

        num_levels = layer.num_octree_levels

        # Show AUTO followed by the valid layer numbers we can choose. AUTO
        # means OctreeImage selects the appropriate octree level
        # dynamically as you zoom in or out. Which is the normal behavior
        # the user almost always wants.
        level_options = {"AUTO": -1}
        level_options.update({str(x): x for x in np.arange(0, num_levels)})
        self.level = QtLabeledComboBox(
            "Octree Level", level_options, on_set_level
        )
        self.addWidget(self.level)

        # Delay for debugging (simulates latency).
        self.mean_combo = _create_mean_combo(on_set_delay)
        self.addWidget(self.mean_combo)
        self.std_dev_combo = _create_std_dev_combo(on_set_delay)
        self.addWidget(self.std_dev_combo)

        # Show some keys and values about the octree.
        self.table = QtSimpleTable()
        self.addWidget(self.table)

        self.set_controls(layer)  # Initial settings.

    def _create_checkbox(self, label, initial_value, callback):
        checkbox = QCheckBox(label)
        checkbox.stateChanged.connect(callback)
        checkbox.setChecked(initial_value)
        self.addWidget(checkbox)

    def set_controls(self, layer: OctreeImage):
        """Set controls based on the layer.

        Parameters
        ----------
        layer : OctreeImage
            Set controls based on this layer.
        """
        self.level.set_value(
            "AUTO" if not layer.freeze_level else layer.octree_level
        )
        self.table.set_values(_get_table_values(layer))


class QtOctreeInfo(QFrame):
    """Frame showing the octree level and tile size.

    layer : OctreeImage
        Show info about this layer.
    """

    def __init__(self, layer: OctreeImage):
        super().__init__()
        self.layer = layer

        self.layout = QtOctreeInfoLayout(layer)
        self.setLayout(self.layout)

        # Initial update and connect for future updates.
        self._update()  # initial update
        layer.events.freeze_level.connect(self._update)
        layer.events.octree_level.connect(self._update)
        layer.events.tile_size.connect(self._update)

    def _update(self, _event=None):
        """Set controls based on the current layer setting."""
        self.layout.set_controls(self.layer)



"""AsyncConfig and OctreeConfig.

Async/octree has its own little JSON config file. This is temporary because
napari does not yet its own system-wide config file. Once that config file
exists, we can roll these settings into there.
"""
import errno
import json
import logging
import os
from collections import namedtuple
from pathlib import Path

from ....utils import config

LOGGER = logging.getLogger("napari.async")

# NAPARI_ASYNC=0 will use these settings, although currently with async
# in experimental this module will not even be imported if NAPARI_ASYNC=0.
DEFAULT_SYNC_CONFIG = {"synchronous": True}

# NAPARI_ASYNC=1 will use these default settings:
DEFAULT_ASYNC_CONFIG = {
    "log_path": None,
    "synchronous": False,
    "num_workers": 6,
    "use_processes": False,
    "auto_sync_ms": 30,
    "delay_queue_ms": 100,
}

# The async config settings.
AsyncConfig = namedtuple(
    "AsyncConfig",
    [
        "log_path",
        "synchronous",
        "num_workers",
        "use_processes",
        "auto_sync_ms",
        "delay_queue_ms",
        "octree",
    ],
)

# The octree config settings.
OctreeConfig = namedtuple("OctreeConfig", ["tile_size"],)

DEFAULT_OCTREE_CONFIG = {"tile_size": 256}


def _log_to_file(path: str) -> None:
    """Log "napari.async" messages to the given file.

    Parameters
    ----------
    path : str
        Log to this file path.
    """
    if path:
        fh = logging.FileHandler(path)
        LOGGER.addHandler(fh)
        LOGGER.setLevel(logging.DEBUG)


def _load_config(config_path: str) -> dict:
    """Load the JSON formatted config file.

    Parameters
    ----------
    config_path : str
        The path of the JSON file we should load.

    Return
    ------
    dict
        The parsed data from the JSON file.
    """
    path = Path(config_path).expanduser()
    if not path.exists():
        # Produce a nice error message like:
        #     Config file NAPARI_ASYNC=missing-file.json not found
        raise FileNotFoundError(
            errno.ENOENT, f"Config file NAPARI_ASYNC={path} not found", path,
        )

    with path.open() as infile:
        return json.load(infile)


def _get_config_data() -> dict:
    """Return the config data from the user's file or the default data.

    Return
    ------
    dict
        The config data we should use.
    """
    value = os.getenv("NAPARI_ASYNC")

    if (value is None or value == "0") and not config.async_octree:
        # Async is disabled. Note, while async is experimental if
        # NAPARI_ASYNC and NAPARI_OCTREE are both not set, then actually
        # this code will not be run. No async related code is even
        # imported.
        #
        # However long term, using DEFAULT_SYNC_CONFIG means the
        # ChunkLoader is being used, but ChunkLoader.synchronous
        # is set True. Meaning every single load is synchronous.
        #
        # Once the feature is not experimental that will probably
        # be the way to turn async "off". It will still run through
        # the ChunkLoader, but the loads will be synchronous.
        return DEFAULT_SYNC_CONFIG

    # Async is enabled with defaults.
    async_defaults = DEFAULT_ASYNC_CONFIG
    async_defaults['octree'] = DEFAULT_OCTREE_CONFIG
    return async_defaults

    return _load_config(value)  # Load the user's JSON config file.


def _create_async_config(data: dict) -> AsyncConfig:
    """Creates the AsyncConfig named tuple and set up logging.

    Parameters
    ----------
    data : dict
        The config settings.

    Return
    ------
    AsyncConfig
        The config settings to use.
    """
    octree_data = data.get("octree")
    if octree_data is None:
        octree_data = DEFAULT_OCTREE_CONFIG

    octree_config = OctreeConfig(tile_size=octree_data.get("tile_size", 64))

    config = AsyncConfig(
        log_path=data.get("log_path"),
        synchronous=data.get("synchronous", True),
        num_workers=data.get("num_workers", 6),
        use_processes=data.get("use_processes", False),
        auto_sync_ms=data.get("auto_sync_ms", 30),
        delay_queue_ms=data.get("delay_queue_ms", 100),
        octree=octree_config,
    )

    _log_to_file(config.log_path)
    LOGGER.info("_create_async_config = ")
    LOGGER.info(json.dumps(data, indent=4, sort_keys=True))

    return config


# The global config settings instance.
async_config = _create_async_config(_get_config_data())



"""Functions to downsample images and create multi-scale images.

This file is early/placeholder. In normal useage we might never create
tiles, because downsampling images is very slow. But for debugging and
development we do make tiles on the fly, for test images and other reasons.

Long term we might possible make tiles in the background at some point. So
as you browse a large image that doesn't have tiles, they are created in
the background. But that's pretty speculative and far out.
"""
import time
from typing import List

import dask
import dask.array as da
import numpy as np
from scipy import ndimage as ndi

from ....types import ArrayLike
from .octree_util import NormalNoise

TileArray = List[List[ArrayLike]]


def _get_tile(tiles: TileArray, row, col):
    try:
        return tiles[row][col]
    except IndexError:
        return None


def _none(items):
    return all(x is None for x in items)


def _add_delay(array, delay_ms: NormalNoise):
    """Add a random delay when this array is first accessed.

    Parameters
    ----------
    noise : NormalNoise
        The amount of the random delay in milliseconds.
    """

    @dask.delayed
    def delayed(array):
        sleep_ms = max(0, np.random.normal(delay_ms.mean, delay_ms.std_dev))
        time.sleep(sleep_ms / 1000)
        return array

    return da.from_delayed(delayed(array), array.shape, array.dtype)


def _combine_tiles(*tiles: np.ndarray) -> np.ndarray:
    """Combine between one and four tiles into a single tile.

    The single resulting tile is not downsampled, so its size is the size
    of the four tiles combined together. However, typically the result will
    be downsampled by half in the following steps.

    Parameters
    ----------
    tiles
        The 4 child tiles, some might be None.
    """
    if len(tiles) != 4:
        raise ValueError("Must have 4 values")

    if tiles[0] is None:
        raise ValueError("Position 0 cannot be None")

    # The layout of the children is:
    # 0 1
    # 2 3
    if _none(tiles[1:4]):
        # We only have one tile:
        # 0 X
        # X X
        return tiles[0]
    if _none(tiles[2:4]):
        # We only have the top two tiles:
        # 0 1
        # X X
        return np.hstack(tiles[0:2])
    if _none((tiles[1], tiles[3])):
        # We only have the left two tiles:
        # 0 X
        # 2 X
        return np.vstack((tiles[0], tiles[2]))

    # We have all four tiles:
    # 0 1
    # 2 3
    row1 = np.hstack(tiles[0:2])
    row2 = np.hstack(tiles[2:4])
    return np.vstack((row1, row2))


def _create_downsampled_tile(*tiles: np.ndarray) -> np.ndarray:
    """Create one parent tile from four child tiles.

    Parameters
    ----------
    tiles
        The 4 child tiles, some could be None.
    """
    # Combine 1-4 tiles together.
    combined_tile = _combine_tiles(*tiles)

    # Down sample by half.
    return ndi.zoom(
        combined_tile, [0.5, 0.5, 1], mode='nearest', prefilter=True, order=1
    )


def _create_coarser_level(tiles: TileArray) -> TileArray:
    """Return a level that is one level coarser.

    Combine each 2x2 group of tiles into one downsampled tile. This is slow
    so currently it's only used for testing. Most multi-scale data will
    be provided pre-downsampled into multiple levels.

    Parameters
    ----------
    tiles : TileArray
        The tiles to combine.

    Returns
    -------
    TileArray
        The coarser level of tiles.
    """

    level = []

    for row in range(0, len(tiles), 2):
        row_tiles = []
        for col in range(0, len(tiles[row]), 2):
            # The layout of the children is:
            # 0 1
            # 2 3
            group = (
                _get_tile(tiles, row, col),
                _get_tile(tiles, row, col + 1),
                _get_tile(tiles, row + 1, col),
                _get_tile(tiles, row + 1, col + 1),
            )
            tile = _create_downsampled_tile(*group)
            row_tiles.append(tile)
        level.append(row_tiles)

    return level


def create_downsampled_levels(
    image: np.ndarray, tile_size: int
) -> List[np.ndarray]:
    """Return a list of levels coarser then this own.

    The first returned level is half the size of the input image, and each
    additional level is half as small again. The longest size in the
    last level is equal to or smaller than tile_size.

    For example if the tile_size is 256, the data in the file level will
    be smaller than (256, 256).

    Parameters
    ----------
    image : np.darray
        The full image to create levels from.

    Return
    ------
    List[np.ndarray]
        A list of levels where levels[0] is the first downsampled level.
    """
    if image.ndim == 2:
        zoom = [0.5, 0.5]
    else:
        assert image.ndim == 3
        zoom = [0.5, 0.5, 1]

    previous = image
    levels = []

    # Repeat until we have level that will fit in a single tile, that will
    # be come the root/highest level.
    while max(previous.shape) > tile_size:
        next_level = ndi.zoom(
            previous, zoom, mode='nearest', prefilter=True, order=1
        )
        levels.append(next_level)
        previous = levels[-1]

    return levels


def create_multi_scale_from_image(
    image: np.ndarray, tile_size: int
) -> List[np.ndarray]:
    """Turn an image into a multi-scale image with levels.

    Parameters
    ----------
    image : np.darray
        The full image to create levels from.

    Return
    ------
    List[np.ndarray]
        A list of levels where levels[0] is the input image.
    """
    levels = [image] + create_downsampled_levels(image)
    return levels



"""
These convenience functions will be useful for searching pypi for packages
that match the plugin naming convention, and retrieving related metadata.
"""
import configparser
import json
import os
import re
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import lru_cache
from typing import Dict, Generator, List, NamedTuple, Optional, Tuple
from urllib import error, request

PYPI_SIMPLE_API_URL = 'https://pypi.org/simple/'

setup_py_entrypoint = re.compile(
    r"entry_points\s?=\s?([^}]*napari.plugin[^}]*)}"
)
setup_py_pypi_name = re.compile(
    r"setup\s?\(.*name\s?=\s?['\"]([^'\"]+)['\"]", re.DOTALL
)


class ProjectInfo(NamedTuple):
    """Info associated with a PyPI Project."""

    name: str
    version: str
    url: str
    summary: str
    author: str
    license: str


@lru_cache(maxsize=128)
def get_packages_by_prefix(prefix: str) -> Dict[str, str]:
    """Search for packages starting with ``prefix`` on pypi.

    Packages using naming convention: http://bit.ly/pynaming-convention
    can be autodiscovered on pypi using the SIMPLE API:
    https://www.python.org/dev/peps/pep-0503/

    Returns
    -------
    dict
        {name: url} for all packages at pypi that start with ``prefix``
    """

    with request.urlopen(PYPI_SIMPLE_API_URL) as response:
        html = response.read().decode()

    return {
        name: PYPI_SIMPLE_API_URL + url
        for url, name in re.findall(
            f'<a href="/simple/(.+)">({prefix}.*)</a>', html
        )
    }


@lru_cache(maxsize=128)
def get_package_versions(name: str) -> List[str]:
    """Get available versions of a package on pypi

    Parameters
    ----------
    name : str
        name of the package

    Returns
    -------
    tuple
        versions available on pypi
    """
    with request.urlopen(PYPI_SIMPLE_API_URL + name) as response:
        html = response.read()

    return re.findall(f'>{name}-(.+).tar', html.decode())


@lru_cache(maxsize=1)
def get_napari_plugin_repos_from_github() -> List[dict]:
    """Return GitHub API hits for repos with "napari plugin" in the README."""
    with request.urlopen(
        'https://api.github.com/search/repositories?q="napari+plugin"+in:readme'
    ) as response:
        return json.loads(response.read().decode()).get("items")


@lru_cache(maxsize=128)
def ensure_published_at_pypi(
    name: str, min_dev_status=3
) -> Optional[ProjectInfo]:
    """Return name if ``name`` is a package in PyPI with dev_status > min."""
    try:
        with request.urlopen(f'https://pypi.org/pypi/{name}/json') as resp:
            info = json.loads(resp.read().decode()).get("info")
    except error.HTTPError:
        return None
    classifiers = info.get("classifiers")
    for i in range(1, min_dev_status):
        if any(f'Development Status :: {1}' in x for x in classifiers):
            return None

    return ProjectInfo(
        name=normalized_name(info["name"]),
        version=info["version"],
        url=info["home_page"],
        summary=info["summary"],
        author=info["author"],
        license=info["license"] or "UNKNOWN",
    )


def ensure_repo_is_napari_plugin(repo_info: dict) -> Optional[ProjectInfo]:
    """Return ProjectInfo of published napari plugin or None.

    This function looks for a setup.py or setup.cfg file in the default branch
    of the repo, then looks for a "napari.plugins" in the entry_points section.
    As such, it will only currently find projects that use setuptools.

    NOTE: this is a hack because we don't have a trove classifier on PyPI (and
    may not be able to get one)

    Parameters
    ----------
    repo_info : dict
        info as returned from the github API (or
        ``get_napari_plugin_repos_from_github``)
        see: https://developer.github.com/v3/repos/

    Returns
    -------
    info : ProjectInfo, optional
        named tuple with project info or None

    """
    # assume repos starting with napari are following naming convention
    if repo_info['name'].lower().startswith("napari"):
        return ensure_published_at_pypi(repo_info['name'])

    # otherwise... we have to look for the entry_point
    raw_url = 'https://raw.githubusercontent.com/{}/{}/'
    base = raw_url.format(repo_info['full_name'], repo_info['default_branch'])

    # first check setup.py
    try:
        with request.urlopen(base + "setup.py") as resp:
            text = resp.read().decode()
        match = setup_py_entrypoint.search(text)
        if match:
            if any(
                'napari.plugin' in line.split("#")[0]
                for line in match.groups()[0].splitlines()
            ):
                name = setup_py_pypi_name.search(text)
                if name:
                    return ensure_published_at_pypi(name.groups()[0])
    except error.HTTPError:  # usually 404
        pass

    # then check setup.cfg
    try:
        with request.urlopen(base + "setup.cfg") as resp:
            text = resp.read().decode()
        parser = configparser.ConfigParser()
        parser.read_string(text)
        if parser.has_option('options.entry_points', 'napari.plugin'):
            return ensure_published_at_pypi(parser.get("metadata", "name"))
    except error.HTTPError:
        pass
    return None


def iter_napari_plugin_info(
    skip={'napari-plugin-engine'},
) -> Generator[ProjectInfo, None, None]:
    """Return a generator that yields ProjectInfo of available napari plugins.

    By default, requires that packages are at least "Alpha" stage of
    development.  to allow lower, change the ``min_dev_status`` argument to
    ``ensure_published_at_pypi``.
    """
    already_yielded = set()
    with ThreadPoolExecutor(max_workers=8) as executor:
        futures = [
            executor.submit(ensure_published_at_pypi, name)
            for name in get_packages_by_prefix("napari-")
            if name not in skip
        ]
        futures.extend(
            [
                executor.submit(ensure_repo_is_napari_plugin, repo_info)
                for repo_info in get_napari_plugin_repos_from_github()
                if repo_info['name'] not in skip  # repo may not have pkg name
            ]
        )
        for future in as_completed(futures):
            info = future.result()
            if info and info not in already_yielded:
                already_yielded.add(info)
                yield info


@lru_cache(maxsize=1)
def list_outdated() -> Dict[str, Tuple[str, ...]]:
    # slow!
    import subprocess

    from ..utils._appdirs import user_site_packages

    env = os.environ.copy()
    combined = os.pathsep.join(
        [user_site_packages(), env.get("PYTHONPATH", "")]
    )
    env['PYTHONPATH'] = combined
    result = subprocess.check_output(
        [sys.executable, '-m', 'pip', 'list', '--outdated'], env=env
    )
    lines = result.decode().splitlines()
    if len(lines) <= 2:
        return {}
    out = dict()
    for line in lines[2:]:
        name, *rest = line.split()
        out[name] = tuple(rest[:2])
    return out


def normalized_name(name) -> str:
    return re.sub(r"[-_.]+", "-", name).lower()



"""OctreeImage class.
"""
import logging
from typing import List

import numpy as np

from ....components.experimental.chunk import (
    ChunkRequest,
    async_config,
    chunk_loader,
)
from ....utils.events import Event
from ..image import Image
from ._chunked_slice_data import ChunkedSliceData
from ._octree_multiscale_slice import OctreeMultiscaleSlice, OctreeView
from .octree_chunk import OctreeChunk, OctreeChunkKey
from .octree_intersection import OctreeIntersection
from .octree_level import OctreeLevelInfo
from .octree_util import NormalNoise, SliceConfig

LOGGER = logging.getLogger("napari.async.octree")


class OctreeImage(Image):
    """OctreeImage layer.

    Experimental variant of Image that renders using an Octree.

    Intended to eventually replace Image.
    """

    def __init__(self, *args, **kwargs):
        self._tile_size = async_config.octree.tile_size

        # Is this the same as Image._data_level? Which should we use?
        self._octree_level = None

        self._view: OctreeView = None

        self._freeze_level = False
        self._track_view = True
        self._slice = None

        self._show_grid = True

        # Temporary to implement a disabled cache.
        self._last_visible_set = set()

        # For logging only
        self.frame_count = 0

        # For debugging and demos, inject a random delay in from of every
        # octree chunk that we access. To simulate latency from IO or
        # computation.
        self._delay_ms = NormalNoise()

        super().__init__(*args, **kwargs)
        self.events.add(
            freeze_level=Event, octree_level=Event, tile_size=Event
        )

    def _get_value(self):
        """Override Image._get_value()."""
        return (0, (0, 0))  # Fake for now until have octree version.

    @property
    def loaded(self):
        """Has the data for this layer been loaded yet."""
        # TODO_OCTREE: what here?
        return True

    @property
    def _empty(self) -> bool:
        return False  # TODO_OCTREE: what here?

    def _update_thumbnail(self):
        # TODO_OCTREE: replace Image._update_thumbnail with nothing for
        # the moment until we decide how to do thumbnail.
        pass

    @property
    def _data_view(self):
        """Viewable image for the current slice. (compatibility)"""
        # Override Image._data_view
        return np.zeros((64, 64, 3))  # fake: does octree need this?

    @property
    def track_view(self) -> bool:
        """Return True if we changing what's dispays as the view changes.

        Return
        ------
        bool
            True if we are tracking the current view.
        """
        return self._track_view

    @track_view.setter
    def track_view(self, value: bool) -> None:
        """Set whether we are tracking the current view.

        Parameters
        ----------
        value : bool
            True if we should track the current view.
        """
        self._track_view = value

    @property
    def tile_size(self) -> int:
        """Return the edge length of single tile, for example 256.

        Return
        ------
        int
            The edge length of a single tile.
        """
        return self._tile_size

    @property
    def tile_shape(self) -> tuple:
        """Return the shape of a single tile, for example 256x256x3.

        Return
        ------
        tuple
            The shape of a single tile.
        """
        # TODO_OCTREE: Must be an easier way to get this shape based on
        # information already stored in Image class?
        if self.multiscale:
            init_shape = self.data[0].shape
        else:
            init_shape = self.data.shape

        tile_shape = (self.tile_size, self.tile_size)

        if self.rgb:
            # Add the color dimension (usually 3 or 4)
            tile_shape += (init_shape[-1],)

        return tile_shape

    @tile_size.setter
    def tile_size(self, tile_size: int) -> None:
        self._tile_size = tile_size
        self.events.tile_size()
        self._slice = None  # For now must explicitly delete it
        self.refresh()  # Create a new slice.

    @property
    def slice_config(self) -> SliceConfig:
        """Return information about the current octree.

        Return
        ------
        SliceConfig
            Configuration information.
        """
        if self._slice is None:
            return None
        return self._slice.slice_config

    @property
    def octree_level_info(self) -> OctreeLevelInfo:
        """Return information about the current level of the current octree.

        Returns
        -------
        OctreeLevelInfo
            Information about the current octree level.
        """
        if self._slice is None:
            return None
        return self._slice.octree_level_info

    @property
    def freeze_level(self) -> bool:
        """Return True if we are forzen viewing a single octree level.

        When viewing the octree normally, freeze_level is always False, but
        during debugging or other special situations it might be on.

        Returns
        -------
        bool
            True if the view is currently frozen viewing on level.
        """
        return self._freeze_level

    @freeze_level.setter
    def freeze_level(self, freeze: bool) -> None:
        """Set whether we are frozen viewing a single octree level.

        Parameters
        ----------
        value : bool
            True if we should determine the octree level automatically.
        """
        self._freeze_level = freeze
        self.events.freeze_level()

    @property
    def octree_level(self):
        """Return the currently displayed octree level."""
        return self._octree_level

    @octree_level.setter
    def octree_level(self, level: int):
        """Set the octree level we should be displaying.

        Parameters
        ----------
        level : int
            Display this octree level.
        """
        assert 0 <= level < self.num_octree_levels
        self._octree_level = level
        self.events.octree_level()
        if self._slice is not None:
            self._slice.octree_level = level
        self.events.loaded()  # redraw

    @property
    def num_octree_levels(self) -> int:
        """Return the total number of octree levels."""
        return len(self.data)  # Multiscale

    def _new_empty_slice(self) -> None:
        """Initialize the current slice to an empty image.

        Overides Image._new_empty_slice() and does nothing because we don't
        need an empty slice. We create self._slice when
        self._set_view_slice() is called.

        The empty slice was needed to satisfy the old VispyImageLayer that
        used a single ImageVisual. But OctreeImage is drawn with
        VispyTiledImageVisual. It does not need an empty image. It gets
        chunks from our self.visible_chunks property, and it will just draw
        nothing if that returns an empty list.

        When OctreeImage become the only image class, this can go away.
        """

    @property
    def visible_chunks(self) -> List[OctreeChunk]:
        """Chunks in the current slice which in currently in view."""
        if self._slice is None or self._view is None:
            return []

        chunks = self._slice.get_visible_chunks(self._view)

        LOGGER.debug(
            "OctreeImage.visible_chunks: frame=%d num_chunks=%d",
            self.frame_count,
            len(chunks),
        )
        self.frame_count += 1

        visible_set = set(octree_chunk.key for octree_chunk in chunks)

        # Remove any chunks from our self._last_visible set which are no
        # longer in view.
        for key in list(self._last_visible_set):
            if key not in visible_set:
                self._last_visible_set.remove(key)

        # If we switched to a new octree level, update our currently shown level.
        slice_level = self._slice.octree_level
        if self._octree_level != slice_level:
            self._octree_level = slice_level
            self.events.octree_level()

        def _log(i, count, label, chunk):
            LOGGER.debug(
                "Visible Chunk: %d of %d -> %s: %s", i, count, label, chunk
            )

        visible_chunks = []  # TODO_OCTREE combine list/set
        visible_set = set()
        for i, octree_chunk in enumerate(chunks):

            if not chunk_loader.cache.enabled:
                new_in_view = octree_chunk.key not in self._last_visible_set
                if new_in_view and octree_chunk.in_memory:
                    # Not using cache, so if this chunk just came into view
                    # clear it out, so it gets reloaded.
                    octree_chunk.clear()

            if octree_chunk.in_memory:
                # The chunk is fully in memory, we can view it right away.
                # _log(i, len(chunks), "ALREADY LOADED", octree_chunk)
                visible_chunks.append(octree_chunk)
                visible_set.add(octree_chunk.key)
            elif octree_chunk.loading:
                # The chunk is being loaded, do not view it yet.
                _log(i, len(chunks), "LOADING:", octree_chunk)
            else:
                # The chunk is not in memory and is not being loaded, so
                # we are going to load it.
                sync_load = self._load_chunk(octree_chunk)
                if sync_load:
                    # The chunk was loaded synchronously. Either it hit the
                    # cache, or it's fast-loading data. We can draw it now.
                    _log(i, len(chunks), "SYNC LOAD", octree_chunk)
                    visible_chunks.append(octree_chunk)
                    visible_set.add(octree_chunk.key)
                else:
                    # An async load was initiated, sometime later our
                    # self._on_chunk_loaded method will be called.
                    _log(i, len(chunks), "ASYNC LOAD", octree_chunk)

        # Update our _last_visible_set with what is in view.
        for octree_chunk in chunks:
            self._last_visible_set.add(octree_chunk.key)

        return visible_chunks

    def _load_chunk(self, octree_chunk: OctreeChunk) -> None:

        indices = np.array(self._slice_indices)
        key = OctreeChunkKey(self, indices, octree_chunk.location)

        chunks = {'data': octree_chunk.data}

        octree_chunk.loading = True

        # Create the ChunkRequest and load it with the ChunkLoader.
        request = chunk_loader.create_request(self, key, chunks)

        satisfied_request = chunk_loader.load_chunk(request)

        if satisfied_request is None:
            return False  # Load was async.

        # Load was sync so we can insert the data into the octree
        # and we will draw it this frame.
        octree_chunk.data = satisfied_request.chunks.get('data')
        return True

    def _on_data_loaded(self, data: ChunkedSliceData, sync: bool) -> None:
        """The given data a was loaded, use it now."""

    def _update_draw(self, scale_factor, corner_pixels, shape_threshold):

        # Need refresh if have not been draw at all yet.
        # TODO_OCTREE: why? do we really?
        need_refresh = self._view is None

        super()._update_draw(scale_factor, corner_pixels, shape_threshold)

        # Compute our 2D corners from the incoming n-d corner_pixels
        # TODO_OCTREE: Throw in the two copies to fix weird bug, need
        # to fix it for real soon.
        data_corners = (
            self._transforms[1:].simplified.inverse(corner_pixels).copy()
        )
        corners = data_corners[:, self._dims.displayed].copy()

        # Update our self._view to to catpure the state of things right
        # before we are drawn. Our self._view will used by our
        # visible_chunks() method.
        self._view = OctreeView(
            corners, shape_threshold, self.freeze_level, self.track_view
        )

        if need_refresh:
            self.refresh()

    def get_intersection(self) -> OctreeIntersection:
        """The the interesection between the current view and the octree.

        Returns
        -------
        OctreeIntersection
            The intersection between the current view and the octree.
        """
        if self._slice is None:
            return None

        return self._slice.get_intersection(self._view)

    def _outside_data_range(self, indices) -> bool:
        """Return True if requested slice is outside of data range.

        Return
        ------
        bool
            True if requested slice is outside data range.
        """

        extent = self._extent_data
        not_disp = self._dims.not_displayed

        return np.any(
            np.less(
                [indices[ax] for ax in not_disp],
                [extent[0, ax] for ax in not_disp],
            )
        ) or np.any(
            np.greater(
                [indices[ax] for ax in not_disp],
                [extent[1, ax] for ax in not_disp],
            )
        )

    def _set_view_slice(self):
        """Set the view given the indices to slice with.

        This replaces Image._set_view_slice() entirely. The hope is eventually
        this class OctreeImage becomes Image. And the non-tiled multiscale
        logic in Image._set_view_slice goes away entirely.
        """
        if self._slice is not None:
            # For now bail out so we don't nuke an existing slice which
            # contains an existing octree. Soon we'll need to figure out
            # if we are really changing slices (and need a new octree).
            return

        indices = np.array(self._slice_indices)
        if self._outside_data_range(indices):
            return

        # Indices to get at the data we are currently viewing.
        indices = self._get_slice_indices()

        # TODO_OCTREE: easier way to do this?
        base_shape = self.data[0].shape
        base_shape_2d = [base_shape[i] for i in self._dims.displayed]

        slice_config = SliceConfig(
            base_shape_2d, len(self.data), self._tile_size, self._delay_ms
        )

        # OctreeMultiscaleSlice wants all the levels, but only the dimensions
        # of each level that we are currently viewing.
        slice_data = [level_data[indices] for level_data in self.data]

        self._slice = OctreeMultiscaleSlice(
            slice_data, slice_config, self._raw_to_displayed,
        )

    def _get_slice_indices(self) -> tuple:
        """Get the slice indices including possible depth for RGB."""
        indices = tuple(self._slice_indices)

        if self.rgb:
            indices += (slice(None),)

        return indices

    def on_chunk_loaded(self, request: ChunkRequest) -> None:
        """An asynchronous ChunkRequest was loaded.

        Override Image.on_chunk_loaded() fully.

        Parameters
        ----------
        request : ChunkRequest
            This request was loaded.
        """
        if self._slice.on_chunk_loaded(request):
            # Tell the visual to redraw with this new chunk.
            self.events.loaded()

    @property
    def delay_ms(self) -> NormalNoise:
        """Return the currently configured artificial load delay.

        Return
        ------
        NormalNoise
            The current configured delay.
        """
        return self._delay_ms

    @delay_ms.setter
    def delay_ms(self, delay_ms: NormalNoise):
        """Set the new artificial load delay.

        We sometimes want to simulate latency for debugging or demos.

        Parameters
        ----------
        delay_ms : NormalNoise
            Optional delay to simulate latency.
        """
        self._delay_ms = delay_ms
        self._slice = None  # For now must explicitly delete it
        self.refresh()  # Create a new slice.

    @property
    def show_grid(self) -> bool:
        """True if we are drawing a grid on top of the tiles.

        Return
        ------
        bool
            True if we are drawing a grid on top of the tiles.
        """
        return self._show_grid

    @show_grid.setter
    def show_grid(self, show: bool) -> None:
        """Set whether we should draw a grid on top of the tiles.

        Parameters
        ----------
        show : bool
            True if we should draw a grid on top of the tiles.
        """
        self._show_grid = show
        self.events.loaded()  # redraw



import numpy as np

from ..utils.events import EmitterGroup


class GridCanvas:
    """Grid for canvas.

    Right now the only grid mode that is still inside one canvas with one
    camera, but future grid modes could support multiple canvases.

    Attributes
    ----------
    events : EmitterGroup
        Event emitter group
    enabled : bool
        If grid is enabled or not.
    shape : 2-tuple of int
        Number of rows and columns in the grid. A value of -1 for either or
        both of will be used the row and column numbers will trigger an
        auto calculation of the necessary grid shape to appropriately fill
        all the layers at the appropriate stride.
    stride : int
        Number of layers to place in each grid square before moving on to
        the next square. The default ordering is to place the most visible
        layer in the top left corner of the grid. A negative stride will
        cause the order in which the layers are placed in the grid to be
        reversed.
    """

    def __init__(self, *, shape=(-1, -1), stride=1, enabled=False):

        # Events:
        self.events = EmitterGroup(
            source=self,
            auto_connect=True,
            enabled=None,
            stride=None,
            shape=None,
        )

        self._enabled = enabled
        self._stride = stride
        self._shape = shape

    @property
    def enabled(self):
        """bool: If grid is enabled or not."""
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        self._enabled = enabled
        self.events.enabled()

    @property
    def shape(self):
        """2-tuple of int: Number of rows and columns in the grid."""
        return self._shape

    @shape.setter
    def shape(self, shape):
        self._shape = tuple(shape)
        self.events.shape()

    @property
    def stride(self):
        """int: Number of layers in each grid square."""
        return self._stride

    @stride.setter
    def stride(self, stride):
        self._stride = stride
        self.events.stride()

    def actual_shape(self, nlayers=1):
        """Return the actual shape of the grid.

        This will return the shape parameter, unless one of the row
        or column numbers is -1 in which case it will compute the
        optimal shape of the grid given the number of layers and
        current stride.

        If the grid is not enabled, this will return (1, 1).

        Parameters
        ----------
        nlayers : int
            Number of layers that need to be placed in the grid.

        Returns
        -------
        shape : 2-tuple of int
            Number of rows and columns in the grid.
        """
        if self.enabled:
            n_row, n_column = self.shape
            n_grid_squares = np.ceil(nlayers / abs(self.stride)).astype(int)

            if n_row == -1 and n_column == -1:
                n_column = np.ceil(np.sqrt(n_grid_squares)).astype(int)
                n_row = np.ceil(n_grid_squares / n_column).astype(int)
            elif n_row == -1:
                n_row = np.ceil(n_grid_squares / n_column).astype(int)
            elif n_column == -1:
                n_column = np.ceil(n_grid_squares / n_row).astype(int)

            n_row = max(1, n_row)
            n_column = max(1, n_column)

            return (n_row, n_column)
        else:
            return (1, 1)

    def position(self, index, nlayers):
        """Return the position of a given linear index in grid.

        If the grid is not enabled, this will return (0, 0).

        Parameters
        ----------
        index : int
            Position of current layer in layer list.
        nlayers : int
            Number of layers that need to be placed in the grid.

        Returns
        -------
        position : 2-tuple of int
            Row and column position of current index in the grid.
        """
        if self.enabled:
            n_row, n_column = self.actual_shape(nlayers)

            # Adjust for forward or reverse ordering
            if self.stride < 0:
                adj_i = nlayers - index - 1
            else:
                adj_i = index

            adj_i = adj_i // abs(self.stride)
            adj_i = adj_i % (n_row * n_column)
            i_row = adj_i // n_column
            i_column = adj_i % n_column
            return (i_row, i_column)
        else:
            return (0, 0)



"""QtImageInfo class.
"""
from qtpy.QtWidgets import QFrame, QLabel, QVBoxLayout


def _get_shape(data) -> tuple:
    """Get shape of the data.

    Return
    ------
    tuple
        The shape of the data.
    """
    if isinstance(data, list):
        return data[0].shape  # Level zero of multiscale.
    return data.shape  # Single scale.


class QtImageInfo(QFrame):
    """Frame showing image shape.

    layer : Layer
        Show info about this layer.
    """

    def __init__(self, layer):
        super().__init__()
        layout = QVBoxLayout()
        shape = _get_shape(layer.data)
        layout.addWidget(QLabel(f"Shape: {shape}"))
        self.setLayout(layout)



"""TextureAtlas2D class.

A texture atlas is a large texture that stores many smaller texture tiles.
"""
from typing import NamedTuple, Optional, Tuple

import numpy as np
from vispy.gloo import Texture2D

from ...layers.image.experimental import OctreeChunk

# Two triangles which cover a [0..1, 0..1] quad.
_QUAD = np.array(
    [[0, 0], [1, 0], [1, 1], [0, 0], [1, 1], [0, 1]], dtype=np.float32,
)


def _quad(size: np.ndarray, pos: np.ndarray) -> np.ndarray:
    """Return one quad with the given size and position.

    Parameters
    ----------
    size : np.ndarray
        Size of the quad (X, Y).
    pos : np.ndarray
        Position of the quad (X, Y)
    """
    quad = _QUAD.copy()

    # Modify the copy in place.
    quad[:, :2] *= size
    quad[:, :2] += pos

    return quad


def _chunk_verts(octree_chunk: OctreeChunk) -> np.ndarray:
    """Return a quad for the vertex buffer.

    Parameters
    ----------
    octree_chunk : OctreeChunk
        Create a quad for this chunk.

    Return
    ------
    np.darray
        The quad vertices.
    """
    geom = octree_chunk.geom
    scaled_shape = octree_chunk.data.shape[:2] * geom.scale
    size = scaled_shape[::-1]  # Reverse into (X, Y) form.

    return _quad(size, geom.pos)


class AtlasTile(NamedTuple):
    """Information about one specific tile in the atlas.

    AtlasTile is returned from TextureAtlas2D.add_tile() so the caller has
    the texture coordinates to render each tile in the atlas.
    """

    index: int
    verts: np.ndarray
    tex_coords: np.ndarray


class TileSpec(NamedTuple):
    """Information about the tiles we are using in the atlas."""

    shape: np.ndarray
    ndim: int
    height: int
    width: int
    depth: int

    @classmethod
    def from_shape(cls, shape: np.ndarray):
        """Create a TileInfo from just the shape."""
        ndim = len(shape)
        assert ndim in [2, 3]  # 2D or 2D with color.
        height, width = shape[:2]
        depth = 1 if ndim == 2 else shape[2]
        return cls(shape, ndim, height, width, depth)

    def is_compatible(self, data: np.ndarray) -> bool:
        """Return True if the given data is compatible with our tiles.

        Parameters
        ----------
        data : np.ndarray
            Return True if this data is compatible with our tiles.
        """
        if self.ndim != data.ndim:
            return False  # Different number of dimensions.

        if self.ndim == 3 and self.depth != data.shape[2]:
            return False  # Different depths.

        if data.shape[0] > self.height or data.shape[1] > self.width:
            return False  # Data is too big for the tile.

        # It's either an exact match, or it's compatible but smaller than
        # the full tile, which is fine.
        return True


class TextureAtlas2D(Texture2D):
    """A two-dimensional texture atlas.

    A single large texture with "slots" for smaller texture tiles.

    Parameters
    ----------
    tile_shape : tuple
        The (height, width) of one tile in texels.
    shape_in_tiles : Tuple[int, int]
        The (height, width) of the full texture in terms of tiles.
    """

    # If True we mark removed tiles red for debugging. When we remove a
    # tile we do not need to waste bandwidth clearing the old texture data.
    # We just let that slot get written to later on. However for debugging
    # we can write into that spot to mark it as empty.
    MARK_DELETED_TILES = False

    def __init__(
        self, tile_shape: tuple, shape_in_tiles: Tuple[int, int], **kwargs,
    ):
        # Each tile's shape in texels, for example (256, 256, 3).
        self.spec = TileSpec.from_shape(tile_shape)

        # The full texture's shape in tiles, for example 4x4.
        self.shape_in_tiles = shape_in_tiles

        # The full texture's shape in texels, for example 1024x1024.
        height = self.spec.height * self.shape_in_tiles[0]
        width = self.spec.width * self.shape_in_tiles[1]
        self.full_shape = np.array(
            [width, height, self.spec.depth], dtype=np.int32
        )

        # Total number of texture slots in the atlas.
        self.num_slots_total = shape_in_tiles[0] * shape_in_tiles[1]

        # Every index is free initially.
        self._free_indices = set(range(0, self.num_slots_total))

        # Pre-compute the texture coords for every tile. Otherwise we'd be
        # calculating these over and over as tiles are added. These are for
        # full tiles only. Edge and corner tiles will need custom texture
        # coordinates based on their size.
        #
        # TODO_OCTREE: Put all these into one compact ndarray?
        #
        tile_shape = self.spec.shape  # Use the shape of a full tile.
        self._tex_coords = [
            self._calc_tex_coords(tile_index, tile_shape)
            for tile_index in range(self.num_slots_total)
        ]

        if self.MARK_DELETED_TILES:
            shape = self.spec.shape
            self.deleted_tile_data = np.empty(shape, dtype=np.uint8)
            self.deleted_tile_data[:] = (1, 1, 1)  # handle RGB or RGBA?

        super().__init__(shape=tuple(self.full_shape), **kwargs)

    @property
    def num_slots_free(self) -> int:
        """Return the number of available texture slots.

        Return
        ------
        int
            The number of available texture slots.
        """
        return len(self._free_indices)

    @property
    def num_slots_used(self) -> int:
        """Return the number of used texture slots.

        Return
        ------
        int
            The number of used texture slots.
        """
        return self.num_slots_total - self.num_slots_free

    def _offset(self, tile_index: int) -> Tuple[int, int]:
        """Return the (row, col) offset into the full atlas texture.

        Parameters
        ----------
        tile_index : int
            Get the offset of this tile.

        Return
        ------
        Tuple[int, int]
            The (row, col) offset of this tile in texels.
        """
        width_tiles = self.shape_in_tiles[1]
        try:
            row = int(tile_index / width_tiles)
        except TypeError:
            pass
        col = tile_index % width_tiles

        return row * self.spec.height, col * self.spec.width

    def _calc_tex_coords(
        self, tile_index: int, tile_shape: np.ndarray,
    ) -> np.ndarray:
        """Return the texture coordinates for this tile.

        This is only called from __init__ when we pre-compute the
        texture coordinates for every tiles.

        Parameters
        ----------
        tile_index : int
            Return coordinates for this tile.

        Return
        ------
        np.ndarray
            A (6, 2) array of texture coordinates.
        """
        offset = self._offset(tile_index)
        pos = offset / self.full_shape[:2]
        shape = tile_shape[:2] / self.full_shape[:2]

        pos = pos[::-1]
        shape = shape[::-1]

        return _quad(shape, pos)

    def add_tile(self, octree_chunk: OctreeChunk) -> Optional[AtlasTile]:
        """Add one tile to the atlas.

        Parameters
        ----------
        data : np.ndarray
            The image data for this one tile.
        """
        data = octree_chunk.data
        assert isinstance(data, np.ndarray)

        if not self.spec.is_compatible(data):
            # It will be not compatible of number of dimensions or depth
            # are wrong. Or if the data is too big to fit in one tile.
            raise ValueError(
                f"Data with shape {octree_chunk.data.shape} is not compatible "
                f"with this TextureAtlas2D which has tile shape {self.spec.shape}"
            )

        try:
            tile_index = self._free_indices.pop()
        except KeyError:
            return None  # No available texture slots.

        # Upload the texture data for this one tile.
        self._set_tile_data(tile_index, data)

        # Return AtlasTile. The caller will need the texture coordinates to
        # render quads using our tiles.
        verts = _chunk_verts(octree_chunk)
        tex_coords = self._get_tex_coords(tile_index, data)
        return AtlasTile(tile_index, verts, tex_coords)

    def _get_tex_coords(self, tile_index: int, data: np.ndarray) -> np.ndarray:
        """Return the texture coordinates for this tile.

        Parameters
        ----------
        tile_index : int
            The index of this tile.
        data : np.ndarray
            The image data for this tile.

        Return
        ------
        np.ndarray
            The texture coordinates for the tile.
        """
        # If it's the exact size of our tiles. Return the pre-computed
        # texture coordinates for this tile. Fast!
        if self.spec.shape == data.shape:
            return self._tex_coords[tile_index]

        # It's smaller than a full size tile, so compute exact coords.
        return self._calc_tex_coords(tile_index, data.shape)

    def remove_tile(self, tile_index: int) -> None:
        """Remove a tile from the texture atlas.

        Parameters
        ----------
        tile_index : int
            The index of the tile to remove.
        """
        self._free_indices.add(tile_index)  # This tile_index is now available.

        if self.MARK_DELETED_TILES:
            self._set_tile_data(tile_index, self.deleted_tile_data)

    def _set_tile_data(self, tile_index: int, data: np.ndarray) -> None:
        """Upload the texture data for this one tile.

        Note the data might be smaller than a full tile slot. If so we
        don't really care what texels are in the rest of the tile's slot.
        They will not be drawn because we'll use the correct texture
        coordinates for the small tile.

        We could instead pad this data up to a full tile size. So we always
        upload the same full size. So there is no dead space. But for now
        we send just the exact data and no more. A tiny bit faster.

        Parameters
        ----------
        tile_index : int
            The index of the tile to upload.
        data
            The texture data for the tile.
        """
        # The texel offset of this tile within the larger texture.
        offset = self._offset(tile_index)

        # Texture2D.set_data() will use glTexSubImage2D() under the hood to
        # only write into the tile's portion of the larger texture. This is
        # a big reason adding tiles to TextureAtlas2D is fast.
        self.set_data(data, offset=offset, copy=True)



"""TiledImageVisual class

A visual that draws tiles using a texture atlas.
"""
from typing import List, Set

import numpy as np

from ...layers.image.experimental import OctreeChunk
from ..vendored import ImageVisual
from ..vendored.image import _build_color_transform
from .texture_atlas import TextureAtlas2D
from .tile_set import TileSet

# Shape of she whole texture in tiles. Hardcode for now.
SHAPE_IN_TILES = (16, 16)


class TiledImageVisual(ImageVisual):
    """An image that is drawn using smaller tiles.

    TiledImageVisual draws a single large image using a set of square image
    tiles. The size of the tiles is configurable, but 256x256 or 512x512
    might be good choices. All the tiles in one TiledImageVisual are the
    same size.

    The tiles are stored in larger textures as an "atlas". An atlas is
    basically just a texture which looks like a grid of smaller images. The
    grid has no borders between the tiles. The size of the larger textures
    is also configurable. For example a single 4096x4096 texture could
    store 256 different 256x256 tiles.

    Adding or removing tiles from a TiledImageVisual is efficient. Only the
    bytes in the tile(s) being updated are sent to the card. The Vispy
    method BaseTexture.set_data() has an "offset" argument. When setting
    texture data with an offset under the hood Vispy calls
    glTexSubImage2D(). It will only update the rectangular region within
    the texture that's being updated.

    In addition, uploading new tiles does not cause the shader to be
    rebuilt. This is another reason TiledImageVisual is faster than
    creating a stand-alone ImageVisuals to draw each tile.

    Finally, rendering the tiles is efficient. TiledImageVisual renders by
    drawing one single list of quads. The texture coordinates of the quads
    point to the various tiles in the texture atlas. If all the tiles are
    stored in the same large texture, there will be zero texture swaps,
    which are expensive.

    Parameters
    ----------
    tile_shape : np.ndarray
        The shape of one tile like (256, 256, 3).
    """

    def __init__(self, tile_shape: np.ndarray, *args, **kwargs):
        self.tile_shape = tile_shape

        self._tiles = TileSet()  # The tiles we are drawing.

        self._clim = np.array([0, 1])  # TOOD_OCTREE: need to support clim

        # Initialize our parent ImageVisual.
        super().__init__(*args, **kwargs)

        # Must create the texture atlas after calling __init__ so
        # the attribute self._interpolation exists.
        self.unfreeze()
        self._texture_atlas = self._create_texture_atlas(tile_shape)
        self.freeze()

    def _create_texture_atlas(self, tile_shape: np.ndarray) -> TextureAtlas2D:
        """Create texture atlas up front or if we change texture shape.

        Attributes
        ----------
        tile_shape : np.ndarray
            The shape of our tiles such as (256, 256, 4).

        Return
        ------
        TextureAtlas2D
            The newly created texture atlas.
        """
        interp = 'linear' if self._interpolation == 'bilinear' else 'nearest'
        return TextureAtlas2D(tile_shape, SHAPE_IN_TILES, interpolation=interp)

    def set_data(self, image) -> None:
        """Set data of the ImageVisual.

        VispyImageLayer._on_display_change calls this with an empty image, but
        we can just ignore it. When created we are "empty" by virtue of not
        drawing any tiles yet.
        """

    def set_tile_shape(self, tile_shape: np.ndarray) -> None:
        """Set the shape of our tiles.

        All tiles are the same shape in terms of texels. However they might
        be drawn different physical sizes. For example drawing a single
        view into a quadtree might end up drawing some tiles 2X or 4X
        bigger than others. Typically you want to draw the "best available"
        data which might be on a different level.

        Parameters
        ----------
        tile_shape : np.ndarray
            Our tiles shape like (256, 256, 4)
        """

        # Clear all our previous tile information and set the new shape.
        self._tiles.clear()
        self.tile_shape = tile_shape

        # Create the new atlas and tell the shader about it.
        self._texture_atlas = self._create_texture_atlas(tile_shape)
        self._data_lookup_fn['texture'] = self._texture_atlas

    @property
    def size(self):
        # TODO_OCTREE: need to compute the size...
        #
        # ImageVisual.size() does
        #     return self._data.shape[:2][::-1]
        #
        # We don't have a self._data so what do we put here? Maybe need
        # a bounds for all the currently visible tiles?
        # return self._texture_atlas.texture_shape[:2]
        return (1024, 1024)

    @property
    def num_tiles(self) -> int:
        """Return the number tiles currently being drawn.

        Return
        ------
        int
            The number of tiles currently being drawn.
        """
        return self._texture_atlas.num_slots_used

    @property
    def octree_chunk(self) -> List[OctreeChunk]:
        """Return data for the chunks we are drawing.

        List[OctreeChunk]
            The data for the chunks we are drawing.
        """
        return self._tiles.chunks

    def add_chunks(self, chunks: List[OctreeChunk]):
        """Any any chunks that we are not already drawing.

        Parameters
        ----------
        chunks : List[OctreeChunk]
            Add any of these we are not already drawing.
        """
        for octree_chunk in chunks:
            if not self._tiles.contains_octree_chunk(octree_chunk):
                self.add_one_tile(octree_chunk)

    def add_one_tile(self, octree_chunk: OctreeChunk) -> None:
        """Add one tile to the tiled image.

        Parameters
        ----------
        octree_chunk : OctreeChunk
            The data for the tile we are adding.

        Return
        ------
        int
            The tile's index.
        """

        atlas_tile = self._texture_atlas.add_tile(octree_chunk)

        if atlas_tile is None:
            return  # No slot available in the atlas.

        self._tiles.add(octree_chunk, atlas_tile)
        self._need_vertex_update = True

    def remove_tile(self, tile_index: int) -> None:
        """Remove one tile from the image.

        Parameters
        ----------
        tile_index : int
            The tile to remove.
        """
        try:
            self._tiles.remove(tile_index)
            self._texture_atlas.remove_tile(tile_index)
            self._need_vertex_update = True
        except IndexError:
            raise RuntimeError(f"Tile index {tile_index} not found.")

    def prune_tiles(self, visible_set: Set[OctreeChunk]) -> None:
        """Remove tiles that are not part of the given visible set.

        visible_set : Set[OctreeChunk]
            The set of currently visible chunks.
        """
        for tile_data in list(self._tiles.tile_data):
            if tile_data.octree_chunk.key not in visible_set:
                tile_index = tile_data.atlas_tile.index
                self.remove_tile(tile_index)

    def _build_vertex_data(self) -> None:
        """Build vertex and texture coordinate buffers.

        This overrides ImageVisual._build_vertex_data(), it is called from
        our _prepare_draw().

        This is the heart of tiled rendering. Instead of drawing one quad
        with one texture, we draw one quad per tile. And for each quad its
        texture coordinates will pull from the right slot in the atlas.

        So as the card draws the tiles, where it's sampling from the
        texture will hop around in the atlas texture.
        """
        if len(self._tiles) == 0:
            return  # Nothing to draw.

        verts = np.zeros((0, 2), dtype=np.float32)
        tex_coords = np.zeros((0, 2), dtype=np.float32)

        # TODO_OCTREE: We can probably avoid vstack here if clever,
        # maybe one one vertex buffer sized according to the max
        # number of tiles we expect. But grow if needed.
        for tile_data in self._tiles.tile_data:
            tile = tile_data.atlas_tile
            verts = np.vstack((verts, tile.verts))
            tex_coords = np.vstack((tex_coords, tile.tex_coords))

        # Set the base ImageVisual _subdiv_ buffers
        self._subdiv_position.set_data(verts)
        self._subdiv_texcoord.set_data(tex_coords)
        self._need_vertex_update = False

    def _build_texture(self) -> None:
        """Override of ImageVisual._build_texture().

        TODO_OCTREE: This needs work. Need to do the clim stuff in in the
        base ImageVisual._build_texture but do it for each tile?
        """
        self._clim = np.array([0, 1])

        self._texture_limits = np.array([0, 1])  # hardcode
        self._need_colortransform_update = True

        self._need_texture_upload = False

    def _prepare_draw(self, view) -> None:
        """Override of ImageVisual._prepare_draw()

        TODO_OCTREE: See how much this changes from base class, if we can
        avoid too much duplication. Or factor out some common methods.
        """
        if self._need_interpolation_update:
            # Call the base ImageVisual._build_interpolation()
            self._build_interpolation()

            # But override to use our texture atlas.
            self._data_lookup_fn['texture'] = self._texture_atlas

        # We call our own _build_texture
        if self._need_texture_upload:
            self._build_texture()

        # TODO_OCTREE: how does colortransform change for tiled?
        if self._need_colortransform_update:
            prg = view.view_program
            grayscale = len(self.tile_shape) == 2 or self.tile_shape[2] == 1
            self.shared_program.frag[
                'color_transform'
            ] = _build_color_transform(
                grayscale, self.clim_normalized, self.gamma, self.cmap
            )
            self._need_colortransform_update = False
            prg['texture2D_LUT'] = (
                self.cmap.texture_lut()
                if (hasattr(self.cmap, 'texture_lut'))
                else None
            )

        # We call our own _build_vertex_data()
        if self._need_vertex_update:
            self._build_vertex_data()

        # Call the normal ImageVisual._update_method() unchanged.
        if view._need_method_update:
            self._update_method(view)



"""ChunkedImageLoader classes.
"""
import logging
from typing import Optional

from ....components.experimental.chunk import ChunkKey
from .._image_loader import ImageLoader
from ._chunked_slice_data import ChunkedSliceData

LOGGER = logging.getLogger("napari.async")


class ChunkedImageLoader(ImageLoader):
    """Load images using the Chunkloader: synchronously or asynchronously.

    Attributes
    ----------
    current_key : Optional[ChunkKey]
        The ChunkKey we are currently loading or showing.
    """

    def __init__(self):
        # We're showing nothing to start.
        self.current_key: Optional[ChunkKey] = None

    def load(self, data: ChunkedSliceData) -> bool:
        """Load this ChunkedSliceData (sync or async).

        Parameters
        ----------
        data : ChunkedSliceData
            The data to load

        Return
        ------
        bool
            True if load happened synchronously.
        """
        key = ChunkKey(data.layer, data.indices)
        LOGGER.debug("ChunkedImageLoader.load: %s", key)

        if self.current_key is not None and self.current_key == key:
            # We are already showing this slice, or its being loaded
            # asynchronously. TODO_ASYNC: does this still happen?
            return False

        # Now "showing" this slice, even if it hasn't loaded yet.
        self.current_key = key

        if data.load_chunks(key):
            return True  # Load was sync, load is done.

        return False  # Load was async, so not loaded yet.

    def match(self, data: ChunkedSliceData) -> bool:
        """Return True if slice data matches what we are loading.

        Parameters
        ----------
        data : ChunkedSliceData
            Does this data match what we are loading?

        Return
        ------
        bool
            Return True if data matches.
        """
        key = data.request.key

        if self.current_key == key:
            LOGGER.debug("ChunkedImageLoader.match: accept %s", key)
            return True

        # Probably we are scrolling through slices and we are no longer
        # showing this slice, so drop it. Even if we don't use it, it
        # should get into the cache, so the load wasn't totally wasted.
        LOGGER.debug("ChunkedImageLoader.match: reject %s", key)
        return False



"""OctreeIntersection class.
"""
from typing import List, NamedTuple, Tuple

import numpy as np

from .octree_chunk import OctreeChunk
from .octree_level import OctreeLevel


class OctreeView(NamedTuple):
    """A view into the octree.

    Attributes
    ----------
    corner : np.ndarray
        The two (row, col) corners in data coordinates, base image pixels.
    canvas : np.ndarray
        The shape of the canvas, the window we are drawing into.
    freeze_level : bool
        If True the octree level will not be automatically chosen.
    track_view : bool
        If True which chunks are being rendered should update as the view is moved.
    """

    corners: np.ndarray
    canvas: np.ndarray
    freeze_level: bool
    track_view: bool

    @property
    def data_width(self) -> int:
        """The width between the corners, in data coordinates.

        Return
        ------
            The width in data coordinates.
        """
        return self.corners[1][1] - self.corners[0][1]

    @property
    def auto_level(self) -> bool:
        """True if the octree level should be selected automatically.

        Return
        ------
        bool
            True if the octree level should be selected automatically.
        """
        return not self.freeze_level and self.track_view


class OctreeIntersection:
    """A view's intersection with the octree.

    Parameters
    ----------
    level : OctreeLevel
        The octree level that we intersected with.
    corners_2d : np.ndarray
        The lower left and upper right corners of the view in data coordinates.
    """

    def __init__(self, level: OctreeLevel, view: OctreeView):
        self.level = level

        info = self.level.info

        # TODO_OCTREE: don't split rows/cols so all these pairs of variables
        # are just one variable each? Use numpy more.
        rows, cols = view.corners[:, 0], view.corners[:, 1]

        base = info.slice_config.base_shape

        self.normalized_range = np.array(
            [np.clip(rows / base[0], 0, 1), np.clip(cols / base[1], 0, 1)]
        )

        scaled_rows = rows / info.scale
        scaled_cols = cols / info.scale

        self._row_range = self.row_range(scaled_rows)
        self._col_range = self.column_range(scaled_cols)

    def tile_range(self, span, num_tiles):
        """Return tiles indices needed to draw the span."""

        def _clamp(val, min_val, max_val):
            return max(min(val, max_val), min_val)

        tile_size = self.level.info.slice_config.tile_size

        span_tiles = [span[0] / tile_size, span[1] / tile_size]
        clamped = [
            _clamp(span_tiles[0], 0, num_tiles - 1),
            _clamp(span_tiles[1], 0, num_tiles - 1) + 1,
        ]

        # int() truncates which is what we want
        span_int = [int(x) for x in clamped]
        return range(*span_int)

    def row_range(self, span: Tuple[float, float]) -> range:
        """Return row indices which span image coordinates [y0..y1]."""
        tile_rows = self.level.info.shape_in_tiles[0]
        return self.tile_range(span, tile_rows)

    def column_range(self, span: Tuple[float, float]) -> range:
        """Return column indices which span image coordinates [x0..x1]."""
        tile_cols = self.level.info.shape_in_tiles[1]
        return self.tile_range(span, tile_cols)

    def is_visible(self, row: int, col: int) -> bool:
        """Return True if the tile [row, col] is in the intersection.

        row : int
            The row of the tile.
        col : int
            The col of the tile.
        """

        def _inside(value, value_range):
            return value_range.start <= value < value_range.stop

        return _inside(row, self._row_range) and _inside(col, self._col_range)

    def get_chunks(self, create_chunks=False) -> List[OctreeChunk]:
        """Return chunks inside this intersection.

        Parameters
        ----------
        create_chunks : bool
            If True, create an OctreeChunk at any location that does
            not already have a chunk.
        """
        chunks = []

        # Get every chunk that is within the rectangular region. These are
        # all the chunks we might possibly draw, because they are within
        # the current view.
        #
        # If we've accessed the chunk recently the existing OctreeChunk
        # will be returned, otherwise a new OctreeChunk is created
        # and returned.
        #
        # OctreeChunks can be loaded or unloaded. Unloaded chunks are not
        # drawn until their data as been loaded in. But here we return
        # every chunk within the rectangle.
        for row in self._row_range:
            for col in self._col_range:
                chunk = self.level.get_chunk(
                    row, col, create_chunks=create_chunks
                )
                if chunk is not None:
                    chunks.append(chunk)

        return chunks



"""VispyCanvas class.
"""
from vispy.scene import SceneCanvas

from .utils_gl import get_max_texture_sizes


class VispyCanvas(SceneCanvas):
    """SceneCanvas for our QtViewer class.

    Add two features to SceneCanvas. Ignore mousewheel events with
    modifiers, and get the max texture size in __init__().

    Attributes
    ----------
    max_texture_sizes : Tuple[int, int]
        The max textures sizes as a (2d, 3d) tuple.

    """

    def __init__(self, *args, **kwargs):

        # Since the base class is frozen we must create this attribute
        # before calling super().__init__().
        self.max_texture_sizes = None

        super().__init__(*args, **kwargs)

        # Call get_max_texture_sizes() here so that we query OpenGL right
        # now while we know a Canvas exists. Later calls to
        # get_max_texture_sizes() will return the same results because it's
        # using an lru_cache.
        self.max_texture_sizes = get_max_texture_sizes()

    def _process_mouse_event(self, event):
        """Ignore mouse wheel events which have modifiers."""
        if event.type == 'mouse_wheel' and len(event.modifiers) > 0:
            return
        super()._process_mouse_event(event)



"""TileSet class.
"""
from dataclasses import dataclass
from typing import List

from ...layers.image.experimental import OctreeChunk
from .texture_atlas import AtlasTile


@dataclass
class TileData:
    """Statistics about chunks during the update process."""

    octree_chunk: OctreeChunk  # The data that produced this tile.
    atlas_tile: AtlasTile  # Information from the texture atlas.


class TileSet:
    """The tiles we are drawing.

    Fast test for membership in both directions: dict and a set.
    """

    def __init__(self):
        self._tiles = {}
        self._chunks = set()

    def __len__(self) -> int:
        """Return the number of tiles in the set.

        Return
        ------
        int
            The number of tiles in the set.
        """
        return len(self._tiles)

    def clear(self) -> None:
        """Clear out all our tiles and chunks. Forget everything."""
        self._tiles.clear()
        self._chunks.clear()

    def add(self, octree_chunk: OctreeChunk, atlas_tile: AtlasTile) -> None:
        """Add this TiledData to the set.

        Parameters
        ----------
        tile_data : TileData
            Add this to the set.
        """
        tile_index = atlas_tile.index
        self._tiles[tile_index] = TileData(octree_chunk, atlas_tile)
        self._chunks.add(octree_chunk.key)

    def remove(self, tile_index: int) -> None:
        """Remove the TileData at this index from the set.

        tile_index : int
            Remove the TileData at this index.
        """
        octree_chunk = self._tiles[tile_index].octree_chunk
        del self._tiles[tile_index]
        self._chunks.remove(octree_chunk.key)

    @property
    def chunks(self) -> List[OctreeChunk]:
        """Return all the chunk data that we have.

        Return
        ------
        List[OctreeChunk]
            All the chunk data in the set.
        """
        return [tile_data.octree_chunk for tile_data in self._tiles.values()]

    @property
    def tile_data(self) -> List[TileData]:
        """Return all the tile data in the set.

        Return
        ------
        List[TileData]
            All the tile data in the set.
        """
        return self._tiles.values()

    def contains_octree_chunk(self, octree_chunk: OctreeChunk) -> bool:
        """Return True if the set contains this chunk data.

        Parameters
        ----------
        octree_chunk : OctreeChunk
            Check if OctreeChunk is in the set.

        Return
        ------
        bool
            True if the set contains this chunk data.
        """
        return octree_chunk.key in self._chunks



import numpy as np
import pytest
from vispy.color import Colormap as VispyColormap

from napari.utils.colormaps import Colormap
from napari.utils.colormaps.colormap_utils import (
    AVAILABLE_COLORMAPS,
    ensure_colormap,
    increment_unnamed_colormap,
)


@pytest.mark.parametrize("name", list(AVAILABLE_COLORMAPS.keys()))
def test_colormap(name):
    np.random.seed(0)

    cmap = AVAILABLE_COLORMAPS[name]

    # Test can map random 0-1 values
    values = np.random.rand((50))
    colors = cmap.map(values)
    assert colors.shape == (len(values), 4)

    # Create vispy colormap and check current colormaps match vispy
    # colormap
    vispy_cmap = VispyColormap(*cmap)
    vispy_colors = vispy_cmap.map(values)
    np.testing.assert_almost_equal(colors, vispy_colors, decimal=6)


def test_increment_unnamed_colormap():
    # test that unnamed colormaps are incremented
    names = [
        '[unnamed colormap 0]',
        'existing_colormap',
        'perceptually_uniform',
        '[unnamed colormap 1]',
    ]
    assert increment_unnamed_colormap(names) == '[unnamed colormap 2]'

    # test that named colormaps are not incremented
    named_colormap = 'perfect_colormap'
    assert increment_unnamed_colormap(names, named_colormap) == named_colormap


def test_can_accept_vispy_colormaps():
    """Test that we can accept vispy colormaps."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    vispy_cmap = VispyColormap(colors)
    cmap = ensure_colormap(vispy_cmap)
    assert isinstance(cmap, Colormap)
    np.testing.assert_almost_equal(cmap.colors, colors)


def test_can_accept_napari_colormaps():
    """Test that we can accept napari colormaps."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    napari_cmap = Colormap(colors)
    cmap = ensure_colormap(napari_cmap)
    assert isinstance(cmap, Colormap)
    np.testing.assert_almost_equal(cmap.colors, colors)


def test_can_accept_vispy_colormap_name_tuple():
    """Test that we can accept vispy colormap named type."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    vispy_cmap = VispyColormap(colors)
    cmap = ensure_colormap(('special_name', vispy_cmap))
    assert isinstance(cmap, Colormap)
    np.testing.assert_almost_equal(cmap.colors, colors)
    assert cmap.name == 'special_name'


def test_can_accept_napari_colormap_name_tuple():
    """Test that we can accept napari colormap named type."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    napari_cmap = Colormap(colors)
    cmap = ensure_colormap(('special_name', napari_cmap))
    assert isinstance(cmap, Colormap)
    np.testing.assert_almost_equal(cmap.colors, colors)
    assert cmap.name == 'special_name'


def test_can_accept_named_vispy_colormaps():
    """Test that we can accept named vispy colormap."""
    cmap = ensure_colormap('red')
    assert isinstance(cmap, Colormap)
    assert cmap.name == 'red'


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_can_accept_vispy_colormaps_in_dict():
    """Test that we can accept vispy colormaps in a dictionary."""
    colors_a = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    colors_b = np.array([[0, 0, 0, 1], [1, 0, 0, 1], [0, 0, 1, 1]])
    vispy_cmap_a = VispyColormap(colors_a)
    vispy_cmap_b = VispyColormap(colors_b)
    cmap = ensure_colormap({'a': vispy_cmap_a, 'b': vispy_cmap_b})
    assert isinstance(cmap, Colormap)
    np.testing.assert_almost_equal(cmap.colors, colors_a)
    assert cmap.name == 'a'


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_can_accept_napari_colormaps_in_dict():
    """Test that we can accept vispy colormaps in a dictionary"""
    colors_a = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    colors_b = np.array([[0, 0, 0, 1], [1, 0, 0, 1], [0, 0, 1, 1]])
    napari_cmap_a = Colormap(colors_a)
    napari_cmap_b = Colormap(colors_b)
    cmap = ensure_colormap({'a': napari_cmap_a, 'b': napari_cmap_b})
    assert isinstance(cmap, Colormap)
    np.testing.assert_almost_equal(cmap.colors, colors_a)
    assert cmap.name == 'a'


def test_can_accept_colormap_dict():
    """Test that we can accept vispy colormaps in a dictionary"""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    cmap = ensure_colormap({'colors': colors, 'name': 'special_name'})
    assert isinstance(cmap, Colormap)
    np.testing.assert_almost_equal(cmap.colors, colors)
    assert cmap.name == 'special_name'



"""Utility script to generate copies of icons with colors based
on our themes. Necessary workaround because qt does not allow
for styling svg elements using qss
"""

import os
import re
import shutil
import sys
from subprocess import SubprocessError, check_call
from typing import Dict, List, Tuple

from ..utils.theme import palettes as _palettes

RESOURCES_DIR = os.path.abspath(os.path.dirname(__file__))
SVGPATH = os.path.join(RESOURCES_DIR, 'icons')

svg_tag_open = re.compile(r'(<svg[^>]*>)')


def themify_icons(
    dest_dir: str,
    svg_path: str = SVGPATH,
    palettes: Dict[str, Dict[str, str]] = _palettes,
    color_lookup: Dict[str, str] = None,
) -> List[str]:
    """Create a new "themed" SVG file, for every SVG file in ``svg_path``.

    Parameters
    ----------
    dest_dir : str
        The directory in which to write all of the themed icons.  Individual
        themes will be in subdirectories named after each theme.
    svg_path : str, optional
        The folder to look in for SVG files, by default will search in a folder
        named ``icons`` in the same directory as this file.
    palettes : dict, optional
        A mapping of ``theme_name: theme_dict``, where ``theme_dict`` is a
        mapping of color classes to rgb strings. By default will uses palettes
        from :const:`napari.resources.utils.theme.palettes`.
    color_lookup : dict, optional
        A mapping of icon name to color class.  If the icon name is not in the
        color_lookup, it's color class will be ``"icon"``.

    Returns
    -------
    files : list of str
        a list of generaged SVG filepaths (each relative to dest_dir)
    """

    color_lookup = color_lookup or {
        'visibility': 'text',
        'visibility_off': 'secondary',
        'menu': 'highlight',
        'drop_down': 'secondary',
        'plus': 'secondary',
        'minus': 'secondary',
        'properties_contract': 'secondary',
        'properties_expand': 'secondary',
        'warning': 'warning',
    }
    icon_names = [
        i.replace('.svg', '')
        for i in os.listdir(SVGPATH)
        if i.endswith('.svg')
    ]

    svg_style_insert = r"""<style type="text/css">
    path{fill:{{ color }}}
    polygon{fill:{{ color }}}
    circle{fill:{{ color }}}
    rect{fill:{{ color }}}
    </style>"""

    files = []
    for theme_name, palette in palettes.items():
        palette_dir = os.path.join(dest_dir, theme_name)
        os.makedirs(palette_dir, exist_ok=True)
        for icon_name in icon_names:
            svg_name = icon_name + '.svg'
            new_file = os.path.join(palette_dir, svg_name)
            color = color_lookup.get(icon_name, 'icon')
            css = svg_style_insert.replace('{{ color }}', palette[color])
            with open(os.path.join(SVGPATH, svg_name), 'r') as fr:
                contents = fr.read()
            with open(new_file, 'w') as fw:
                # use regex to find the svg tag and insert css right after
                # (the '\\1' syntax includes the matched tag in the output)
                fw.write(svg_tag_open.sub(f'\\1{css}', contents))
            files.append(os.path.join(theme_name, svg_name))
    return files


def build_resources_qrc(
    dest_dir: str, overwrite: bool = False
) -> Tuple[str, str]:
    """Create a res.qrc file for all icons generated by ``themify_icons``.

    Parameters
    ----------
    dest_dir : str
        The directory in which to write create the output file.  Themed SVG
        icons will also be written to ``dest_dir/themes``.
    overwrite : bool
        Whether to force rebuilding of the icons and res.qrc file, by default
        False

    Returns
    -------
    tuple
        2-tuple of (path-to-qrc.res, path-to-theme-directory).
    """
    qrc_path = os.path.join(dest_dir, 'res.qrc')
    theme_dir = os.path.join(dest_dir, 'themes')
    if os.path.exists(qrc_path) and (not overwrite):
        return qrc_path, theme_dir

    qrc_string = """
    <!DOCTYPE RCC>
    <RCC version="1.0">
    <qresource>
    """
    for filename in themify_icons(theme_dir):
        qrc_string += f'\n    <file>themes/{filename}</file>'

    qrc_string += """
    </qresource>
    </RCC>
    """

    with open(qrc_path, 'w') as f:
        f.write(qrc_string)

    return qrc_path, theme_dir


def _find_rcc_or_raise() -> str:
    """Locate the Qt rcc binary to generate resource files

    1. we always want to use pyrcc5 if it's available, regardless of API
    2. it will sometimes, (if not always) be named pyrcc5.bat on windows...
       but shutil.which() will find that too
    3. We also want to prefer binaries higher up on the path, and we add
       sys.executable to the front of the path (and \\Scripts on windows)
    4. after pyrcc5 we try pyside2-rcc

    see https://github.com/napari/napari/issues/1221
    and https://github.com/napari/napari/issues/1254

    Returns
    -------
    path : str
        Path to the located rcc binary, or None if not found

    Raises
    ------
    FileNotFoundError
        If no executable can be found.
    """
    python_dir = os.path.dirname(sys.executable)
    paths = [python_dir, os.environ.get("PATH", '')]
    if os.name == 'nt':
        paths.insert(0, os.path.join(python_dir, 'Scripts'))
    path = os.pathsep.join(paths)

    for bin_name in ('pyrcc5', 'pyside2-rcc'):
        rcc_binary = shutil.which(bin_name, path=path)
        if rcc_binary:
            yield rcc_binary
    raise FileNotFoundError(
        "Unable to find an executable to build Qt resources (icons).\n"
        "Tried: 'pyrcc5.bat', 'pyrcc5', 'pyside2-rcc'.\n"
        "Please open issue at https://github.com/napari/napari/issues/."
    )


def build_pyqt_resources(out_path: str, overwrite: bool = False) -> str:
    """Build a res.qrc file from icons and convert for python usage.

    calls :func:`build_resources` and then converts using ``pyside2-rcc`` or
    ``pyrcc5`` depending on which is installed in the environment.
    Finally, cleans up autogenerated icon.svgs and res.qrc file after _qt.py
    file is generated

    Parameters
    ----------
    out_path : str
        Path to write the python resource file.
    overwrite : bool, optional
        Whether to force rebuilding of the output file, by default False

    Returns
    -------
    out_path : str
        Path to the python resource file.  Import this file to make the SVGs
        and other resources available to Qt stylesheets.

    References
    ----------
    https://doc.qt.io/qt-5/resources.html
    """

    if os.path.exists(out_path) and not overwrite:
        return out_path

    # build the resource file to the same path
    qrc_path, theme_dir = build_resources_qrc(
        os.path.dirname(out_path), overwrite=overwrite
    )

    # then convert it to a python file
    # When user use pyenv to manage python version it create shortcut
    # to inform in which environment command is available. For example:
    # > pyenv: pyrcc5: command not found
    #
    #   The `pyrcc5' command exists in these Python versions:
    #     3.7.4/envs/napari-pyqt5
    #     napari-pyqt5
    for name in _find_rcc_or_raise():
        try:
            check_call([name, '-o', out_path, qrc_path])
            break
        except SubprocessError:
            pass
    # make sure we import from qtpy
    with open(out_path, "rt") as fin:
        data = fin.read()
        data = data.replace('PySide2', 'qtpy').replace('PyQt5', 'qtpy')
    with open(out_path, "wt") as fin:
        fin.write(data)

    # cleanup.
    # we do this here because pip uninstall napari would not collect these
    # and the final `out_path.py` contains all the necessary bytes info
    shutil.rmtree(theme_dir, ignore_errors=True)
    try:
        os.remove(qrc_path)
    except Exception:
        pass
    return out_path



"""layers.image.experimental
"""
from .octree_chunk import OctreeChunk, OctreeChunkGeom
from .octree_intersection import OctreeIntersection
from .octree_level import OctreeLevel
from .octree_tile_builder import create_multi_scale_from_image
from .octree_util import SliceConfig, TestImageSettings

# from .octree_image import OctreeImage  # circular



"""QtRender widget.
"""


from qtpy.QtWidgets import QVBoxLayout, QWidget

from ....layers.image import Image
from ....layers.image.experimental.octree_image import OctreeImage
from .qt_frame_rate import QtFrameRate
from .qt_image_info import QtImageInfo
from .qt_mini_map import QtMiniMap
from .qt_octree_info import QtOctreeInfo
from .qt_test_image import QtTestImage

SHOW_MINIMAP = False


class QtRender(QWidget):
    """Dockable widget for render controls.

    Parameters
    ----------
    viewer : Viewer
        The napari viewer.
    layer : Optional[Layer]
        Show controls for this layer. If no layer show minimal controls.
    """

    def __init__(self, viewer, layer=None):
        """Create our windgets.
        """
        super().__init__()
        self.viewer = viewer
        self.layer = layer

        layout = QVBoxLayout()

        # Basic info for any image layer.
        if isinstance(layer, Image):
            layout.addWidget(QtImageInfo(layer))

        if isinstance(layer, OctreeImage):
            # Octree specific controls and widgets.
            layout.addWidget(QtOctreeInfo(layer))

            if SHOW_MINIMAP:
                self.mini_map = QtMiniMap(layer)
                layout.addWidget(self.mini_map)

            self.viewer.camera.events.center.connect(self._on_camera_move)

        # Controls to create a new test image.
        layout.addStretch(1)
        layout.addWidget(QtTestImage(viewer))

        # Frame rate meter.
        self.frame_rate = QtFrameRate()
        layout.addWidget(self.frame_rate)

        self.setLayout(layout)

    def _on_camera_move(self, _event=None):
        """Called when the camera was moved."""
        if SHOW_MINIMAP:
            self.mini_map.update()
        self.frame_rate.on_camera_move()



"""OctreeMultiscaleSlice class.

For viewing one slice of a multiscale image using an octree.
"""
import logging
import math
from typing import Callable, List, Optional

import numpy as np

from ....components.experimental.chunk import ChunkRequest
from ....types import ArrayLike
from .._image_view import ImageView
from .octree import Octree
from .octree_chunk import OctreeChunk, OctreeLocation
from .octree_intersection import OctreeIntersection, OctreeView
from .octree_level import OctreeLevelInfo
from .octree_util import SliceConfig

LOGGER = logging.getLogger("napari.async.octree")


class OctreeMultiscaleSlice:
    """View a slice of an multiscale image using an octree."""

    def __init__(
        self,
        data,
        slice_config: SliceConfig,
        image_converter: Callable[[ArrayLike], ArrayLike],
    ):
        self.data = data

        self._slice_config = slice_config

        self._octree = Octree(id(self), data, slice_config)
        self._octree_level = self._octree.num_levels - 1

        thumbnail_image = np.zeros(
            (64, 64, 3)
        )  # blank until we have a real one
        self.thumbnail: ImageView = ImageView(thumbnail_image, image_converter)

    @property
    def octree_level(self) -> int:
        """The current octree level.

        Return
        ------
        int
            The current octree level.
        """
        return self._octree_level

    @octree_level.setter
    def octree_level(self, level: int) -> None:
        """Set the octree level we are viewing.

        Parameters
        ----------
        level : int
            The new level to display.
        """
        self._octree_level = level

    @property
    def loaded(self) -> bool:
        """Return True if the data has been loaded.

        Because octree multiscale is async, we say we are loaded up front even
        though none of our chunks/tiles might be loaded yet.
        """
        return self.data is not None

    @property
    def octree_level_info(self) -> Optional[OctreeLevelInfo]:
        """Return information about the current octree level.

        Return
        ------
        Optional[OctreeLevelInfo]
            Information about current octree level, if there is one.
        """
        if self._octree is None:
            return None
        return self._octree.levels[self._octree_level].info

    def get_intersection(self, view: OctreeView):
        """Return this view's intersection with the octree."""
        assert self._octree
        level = self._get_octree_level(view)
        return OctreeIntersection(level, view)

    def _get_octree_level(self, view: OctreeView):
        index = self._get_octree_level_index(view)
        if index < 0 or index >= self._octree.num_levels:
            raise ValueError(f"Invalid octree level {index}")
        return self._octree.levels[index]

    def _get_octree_level_index(self, view: OctreeView):

        if not view.auto_level:
            # Return current level, do not update it.
            return self._octree_level

        # Find the right level automatically. Choose a level where the texels
        # in the octree tiles are around the same size as screen pixels.
        # We can do this smarter in the future, maybe have some hysterisis
        # so you don't "pop" to the next level as easily, so there is some
        # fudge factor or dead zone.
        ratio = view.data_width / view.canvas[0]

        if ratio <= 1:
            return 0  # Show the best we've got!

        # Choose the right level...
        return min(math.floor(math.log2(ratio)), self._octree.num_levels - 1)

    def get_visible_chunks(self, view: OctreeView) -> List[OctreeChunk]:
        """Return the chunks currently in view.

        Return
        ------
        List[OctreeChunk]
            The chunks which are visible in the given view.
        """
        intersection = self.get_intersection(view)

        if intersection is None:
            return []

        if view.auto_level:
            # Update our self._octree_level based on what level was automatically
            # selected by the intersection.
            self._octree_level = intersection.level.info.level_index

        # Return the chunks in this intersection.
        return intersection.get_chunks(id(self))

    def _get_octree_chunk(self, location: OctreeLocation):
        level = self._octree.levels[location.level_index]
        return level.get_chunk(location.row, location.col)

    def on_chunk_loaded(self, request: ChunkRequest) -> None:
        """An asynchronous ChunkRequest was loaded.

        Override Image.on_chunk_loaded() fully.

        Parameters
        ----------
        request : ChunkRequest
            This request was loaded.
        """
        location = request.key.location
        if location.slice_id != id(self):
            # There was probably a load in progress when the slice was changed.
            # The original load finished, but we are now showing a new slice.
            # Don't consider it error, just ignore the chunk.
            LOGGER.debug("on_chunk_loaded: wrong slice_id: %s", location)
            return False  # Do not load.

        octree_chunk = self._get_octree_chunk(location)
        if not isinstance(octree_chunk, OctreeChunk):
            # This location in the octree is not a OctreeChunk. That's unexpected,
            # becauase locations are turned into OctreeChunk's when a load
            # is initiated. So this is an error, but log it and keep going.
            LOGGER.error(
                "on_chunk_loaded: missing OctreeChunk: %s", octree_chunk
            )
            return False  # Do not load.

        # Looks good, we are loading this chunk.
        LOGGER.debug("on_chunk_loaded: loading %s", octree_chunk)

        # Get the data from the request.
        incoming_data = request.chunks.get('data')

        # Loaded data should always be an ndarray.
        assert isinstance(incoming_data, np.ndarray)

        # Shove the request's ndarray into the octree's OctreeChunk. This octree
        # chunk now has an ndarray as its data, and it can be rendered.
        octree_chunk.data = incoming_data

        # OctreeChunk should no longer need to be loaded. We can probably
        # remove this check eventually, but for now to be sure.
        assert not self._get_octree_chunk(location).needs_load

        return True  # Chunk was loaded.



"""ExperimentalNamespace and CommandProcessor classes.
"""
from .chunk._commands._utils import highlight

HELP_STR = f"""
{highlight("Available Commands:")}
experimental.cmds.loader
"""


class CommandProcessor:
    def __init__(self, layers):
        self.layers = layers

    @property
    def loader(self):
        from .chunk._commands import LoaderCommands

        return LoaderCommands(self.layers)

    def __repr__(self):
        return "Available Commands:\nexperimental.cmds.loader"


class ExperimentalNamespace:
    def __init__(self, layers):
        self.layers = layers

    @property
    def cmds(self):
        return CommandProcessor(self.layers)

    def __repr__(self):
        return "Available Commands:\nexperimental.cmds.loader"
